# -*- coding: utf-8
"""
pdfreactor.parsecfg._configkeys: info about PDFreactor config keys

vim:
%s;^\(\s\+\)\('[^',]\+'\):;\1(\2,):;
"""

def Resource(*args):
    """
    This is needed e.g. for userStyleSheets and integrationStyleSheets; see:
    https://www.pdfreactor.com/product/doc/webservice/python.html#Resource

    >>> Resource('++resource++pdfreactor.plone/export.css')
    {'uri': '++resource++pdfreactor.plone/export.css'}

    For now, we only support a single argument which will be interpreted as the URI.

    >>> Resource('++resource++pdfreactor.plone/export.css', 'more-styles.css')
    Traceback (most recent call last):
      ...
    ValueError: Too many arguments; for now, we only support a single URI!

    """
    if args[1:]:
        raise ValueError('Too many arguments;'
                ' for now, we only support a single URI!')
    return {'uri': args[0]}


_factories = {
        ('disableLinks',): bool,
        ('outputFormat',): dict,
        ('outputFormat', 'type'): str,
        ('outputFormat', 'width'): int,
        ('outputFormat', 'height'): int,
        ('integrationStyleSheets',): list,
        ('integrationStyleSheets', None): Resource,
        ('userStyleSheets',): list,
        ('userStyleSheets', None): Resource,
        }


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8
"""
pdfreactor.parsecfg._parse: Configuration parser for PDFreactor client integrations
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six.moves import range, zip
from six.moves import StringIO

# from six.moves import map, range

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Standard library:
from string import letters
from token import ENDMARKER, NAME, NEWLINE, OP, STRING, tok_name
from tokenize import (
    COMMENT,
    DEDENT,
    INDENT,
    NL,
    TokenError,
    generate_tokens,
    untokenize,
    )

# visaplan:
from visaplan.tools.minifuncs import makeBool
from visaplan.tools.minifuncs import translate_dummy as _
from visaplan.tools.sequences import sequence_slide

# Logging / Debugging:
from visaplan.tools.debug import pp

# from .oldmethods import map2
# from .oldsymbols import *

__all__ = [
    'make_parser',
    ]

SYMBOLS_MAP = {}
_tmp = [('True',  'on', 'yes'),
        ('False', 'off', 'no'),
        ('None',  'null', 'nothing', 'nil'),
        ]
for _tup in _tmp:
    first = _tup[0]
    for name in _tup:
        SYMBOLS_MAP[name.lower()] = first


def make_prefixed(prefix):
    """
    Create a `prefixed` function to prepend symbol names with a prefix:

    >>> f = make_prefixed('PDFreactor')
    >>> f('SomeKnownSymbol')
    'PDFreactor.SomeKnownSymbol'

    Some aliases to important Python builtins, however, are converted to their
    Python names and not prefixed:

    >>> f('on')
    'True'
    >>> f('no')
    'False'
    >>> f('null')
    'None'

    We accept integer numbers:

    >>> f('1')
    '1'

    ... but no floats etc.:

    >>> f('4.2')
    Traceback (most recent call last):
      ...
    ValueError: invalid literal for int() with base 10: '4.2'

    We can choose not to inject a prefix:

    >>> f2 = make_prefixed(None)
    >>> f2('SomeKnownSymbol')
    'SomeKnownSymbol'
    >>> f2('NIL')
    'None'
    >>> f2('1.23')
    Traceback (most recent call last):
      ...
    ValueError: invalid literal for int() with base 10: '1.23'
    """
    if not prefix:
        prefix = ''
    elif not prefix.endswith('.'):
        prefix += '.'
    def prefixed(token):
        if token[0] in '0123456789':
            int(token)
            return token
        ltoken = token.lower()
        if ltoken in SYMBOLS_MAP:
            return SYMBOLS_MAP[ltoken]
        else:
            return prefix + token
    return prefixed


def make_parser(**kwargs):
    """
    Create a parser for PDFreactor client configurations:

    >>> parse = make_parser()

    In the basic case, the parser will understand comment lines;
    it returns a 3-tuple (config, deletions, unused).
    If we don't have special 'control commands', and don't need to apply
    changes to existing configurations, we might be interested in the first
    part only:

    >>> def p0(txt):
    ...     return sorted(parse(txt)[0].items())
    >>> txt = '''# some comment
    ... config.outputFormat = {
    ...     type: OutputType.JPEG,
    ...     width: 640,
    ... }
    ... '''
    >>> parse(txt)
    >>> p0(txt) # doctest: +NORMALIZE_WHITESPACE
    [('type', 'OutputType.JPEG'),
     ('width': 640)]


    """
    pop = kwargs.pop
    prefix = pop('prefix', None)
    commands = pop('commands', None)
    if commands is None:
        commands = set([])
    elif isinstance(commands, str):
        commands = set(commands.split())

    if kwargs:
        invalid = list(kwargs.keys())
        many = bool(invalid[1:])
        raise TypeError('Unsupported option%s: %r%s' % (
                many and 's' or '',
                invalid[0],
                many and ' ...' or '',
                ))

    prefixed = make_prefixed(prefix)

    def parse(txt):
        """
        Return a 3-tuple:

        config -- a dict, as expected e.g. by the PDFreactor.convert method
        deletions -- keys which have been found to be obsolete during
                     processing; not implemented yet.
        unused -- control statements and (suspected) API methods which have not
                  been converted to config dict changes.
        """
        res = {}
        deletions = []
        unused = []
        for statement in gen_restricted_lines(txt, prefixed):
            if isinstance(statement, Assignment):
                key, val = supported_assignment(statement)
                if isinstance(val, dict):
                    dest = res.getdefault(key, {})
                    dest.update(val)
                else:
                    res[key] = val
            elif 0:  # no deletions implemented yet
                pass
            else:
                unused.append(statement)

        return res, deletions, unused

    return parse


def gen_restricted_lines(txt, func=None):
    """
    Generiere "Zeilen" aus dem übergebenen Text:

    >>> txt = '''# ein Kommentar
    ... strict on
    ... setAddLinks(False) # noch ein Kommentar
    ... with_images off
    ... '''

    >>> prefixed = make_prefixed('A')

    >>> list(gen_restricted_lines(txt, prefixed))
    ... # doctest: +NORMALIZE_WHITESPACE
    [<ControlStatement 'strict on '>,
     <ApiCall 'A.setAddLinks (False )'>,
     <ControlStatement 'with_images off '>]

    Einstmals hartcodierte Aufrufe:

    >>> txt2 = '''
    ... setAddLinks(False)
    ... # Enable bookmarks in the PDF document
    ... setAddBookmarks(True)
    ... setCleanupTool(CLEANUP_NONE)
    ... setEncoding('UTF-8')
    ... setJavaScriptMode(JAVASCRIPT_MODE_ENABLED_NO_LAYOUT)
    ... '''
    >>> list(gen_restricted_lines(txt2, prefixed))
    ... # doctest: +NORMALIZE_WHITESPACE
    [<ApiCall 'A.setAddLinks (False )'>,
     <ApiCall 'A.setAddBookmarks (True )'>,
     <ApiCall 'A.setCleanupTool (A.CLEANUP_NONE )'>,
     <ApiCall "A.setEncoding ('UTF-8')">,
     <ApiCall 'A.setJavaScriptMode (A.JAVASCRIPT_MODE_ENABLED_NO_LAYOUT )'>]

    Es können mehrere Anweisungen pro Zeile notiert werden,
    durch Semikolon getrennt:

    >>> txt3 = '; ;setA();setB()'
    >>> list(gen_restricted_lines(txt3, prefixed))
    [<ApiCall 'A.setA ()'>, <ApiCall 'A.setB ()'>]

    Nun zu den Neuerungen zur Unterstützung der neuen PDFreactor-API.
    Als Ersatz für die veralteten API-Aufrufe
    -- die (vorerst?) noch unterstützt und automatisch konvertiert werden --
    unterstützen wir nun direkte Zuweisungen an die config-Variable, die
    der .convert-Methode übergeben wird:

    >>> txt4 = '''
    ... config.outputFormat = {
    ...     type: OutputType.JPEG,
    ...     width: 640,
    ... }
    ... '''
    >>> list(gen_restricted_lines(txt4))      # doctest: +NORMALIZE_WHITESPACE
    [<Assignment 'config .outputFormat
                ={type :OutputType .JPEG ,width :640 ,}'>]
    """
    if not txt:
        return
    buf = []
    alltokens = generate_tokens(StringIO(txt or '').readline)
    for toktup in alltokens:
        ttype, token, startpos, endpos, line = toktup
        if ttype in (NEWLINE, ENDMARKER, DEDENT):
            if buf:
                yield restrictedStatement(buf, func)
                del buf[:]
        elif ttype == OP and token == ';':
            if buf:
                yield restrictedStatement(buf, func)
                del buf[:]
        elif ttype in (NL, COMMENT):
            pass
        elif not buf and ttype in (INDENT,
                ):
            pass
        else:
            buf.append(toktup)
    if buf:
        yield restrictedStatement(buf, func)


# ------------------------ [ Statements, generiert von ApiFilter ... [
class Statement(object):
    """
    Ein von der ApiFilter-Umgebung verwendbares Statement
    """

    def __init__(self, tokens, transform=None):
        tokheads = []
        name = None
        args = []
        has_errors = 0
        for toktup in tokens:
            ttype, token = toktup[:2]
            if name is None:
                if ttype == NAME:
                    name = token
                elif ttype in (INDENT,):
                    continue
                else:
                    if not has_errors:
                        pp(('Hu?', {
                            'tokens:': tokens,
                            'toktup:': toktup[:2],
                            }))
                    token_info(*toktup)
                    print('%s: Name erwartet' % (toktup,))
                    has_errors += 1
            elif ttype not in (OP,
                               ):
                args.append(token)
            # weitere Informationen (toktup[2:]) verwerfen:
            tokheads.append((ttype, token))
        self.tokens = tokheads
        self.name = name
        self.args = args
        if transform is not None:
            self.transformed = True
            tokens2 = []
            for toktup in tokens:
                ttype, token = toktup[:2]
                if ttype == NAME:
                    token = transform(token)
                tokens2.append((ttype, token))

            self.tokens_transformed = tokens2
        else:
            self.transformed = False

    def __str__(self):
        if self.transformed:
            return untokenize(self.tokens_transformed)
        else:
            return untokenize(self.tokens)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))

    def tell(self):
        pp({'name': self.name,
            'args': self.args,
            'tokens': self.tokens,
            'tokens_pretty': resolve_tokens(self.tokens),
            })

class ControlStatement(Statement):
    """
    Zur Steuerung der API-Aufrufe
    """

class ApiCall(Statement):
    """
    API-Aufruf; Methoden- und sonstige Namen werden mit einem Präfix versehen
    """

class Assignment(Statement):
    """
    Assignment of a value to a config key

    To support the new PDFreactor API, which uses a config dictionary instead
    of the plethora of setter methods, we support statements like

      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }

    which would be translated to something like

      dic = config.getdefault('outputFormat', {})
      dic.update({
          'type': PDFreactor.OutputType.JPEG,
          'width': 640,
          })
    """
    def key_and_value(self, skip):
        tokens = self.tokens
        if skip is not None:
            for toktup, chunk in zip(tokens, skip):
                if chunk is None:
                    continue  # accept anything
                elif toktup[1] != chunk:
                    raise ValueError('Expected %(chunk)r; found %(toktup)s!'
                            % locals())
            tokens = tokens[len(skip):]
        if not tokens:
            raise ValueError('Empty assignment: %(self)r' % locals())
        _name = []
        _isdic = None
        _done = 0
        _left_complete = 0
        _braces = []  # a stack of braces
        _thiskey = None
        _thisval = []
        _thedict = None

        for toktup in tokens:
            _typ, _val = toktup
            if _done:
                raise ValueError('Too much: %(toktup)s!' % locals())
            if _typ == OP:
                if not _name:
                    raise ValueError('No name yet!')
                elif _val == '=':
                    if _left_complete:
                        raise ValueError('Duplicate "="!')
                    _left_complete = 1
                elif _val == '{':
                    if not _left_complete:
                        raise ValueError('%(_val)r found but no = yet!'
                                % locals())
                    assert _left_complete
                    if _isdic is not None:
                        raise ValueError('Found %(_val)r,'
                                ' with _isdic == %(_isdic)r'
                                % locals())
                    _isdic = True
                    _thedict = {}
                elif _val == ';':
                    _done = 1
                else:
                    raise ValueError('%(toktup)r not supported yet' % locals())
            elif _left_complete:
                _thisval.append(_val)
            else:
                _name.append(_val)

        return (
            ''.join(_name),
            _thedict if   _isdic
                     else ''.join(_thisval))
# ------------------------ ] ... Statements, generiert von ApiFilter ]


def restrictedStatement(seq, transform_func):
    """
    Erzeuge aus der übergebenen (vom tokenize-Modul erzeugten) Sequenz einen
    API-Aufruf (Klasse ApiCall) oder eine Kontrollanweisung (ControlStatement).

    Einige Symbole werden normalisiert;
    so bestehen z. B. gewisse Freiheiten bei der Angabe logischer Werte:

    >>> func = make_prefixed('A')
    >>> repr(restrictedStatement([(1, 'strict'), (1, 'on')], func))
    "<ControlStatement 'strict on '>"
    >>> restrictedStatement([(1, 'strict'), (1, 'off')], func)
    <ControlStatement 'strict off '>

    Alle anderen Symbole werden, wie auch die Namen der API-Methoden,
    durch das Präfix <prefix> ergänzt:

    >>> restrictedStatement([(1, 'setDings'), (OP, '('),
    ...     (1, 'ATTRNAME'), (OP, ')')], func)
    <ApiCall 'A.setDings (A.ATTRNAME )'>

    (tokenize.untokenize fügt leider nach jedem Namen und jeder Zahl
    sicherheitshalber ein Leerzeichen ein)
    """
    res = []
    main = None
    ltype = None # line type, "Zeilen"typ
    for prevtup, currtup, nexttup in sequence_slide(seq):
        ttype, token = currtup[:2]
        WTF = tok_name[ttype]
        # pp(WTF, token)
        if main is None:
            if ttype == NAME:
                # set_trace()
                nexttype = (None
                            if nexttup is None
                            else nexttup[0])
                if nexttype is OP and nexttup[1] != '(':
                    ltype = CT_ASSIGNMENT
                elif token in CONTROL_COMMANDS:
                    ltype = CT_CONTROL
                elif nexttype == NAME:
                    # we might get an exception during untokenization :-(
                    # ... but at least it's an exception.
                    # We insist in the '.' OP after 'config'!
                    res.append(currtup)
                    raise BogusLine(untokenize(res))
                else:
                    ltype = CT_API
                main = token
        res.append(currtup)
    if ltype == CT_ASSIGNMENT:
        return Assignment(res, transform_func)
    elif ltype == CT_API:
        return ApiCall(res, transform_func)
    elif ltype == CT_CONTROL:
        return ControlStatement(res)
    elif not res:
        return None
    elif ltype is None:
        pp({'ltype:': ltype,
            'main': main,
            'res': res,
            })
        for toktup in res:
            token_info(*toktup)
        raise BogusLine(untokenize(res))
    else:
        raise ProgrammingError
    return ' '.join(res)

API_WHITELIST = frozenset(['enableDebugMode',
                           ])
def acceptable_method_name(name):
    """
    Gehört der übergebene Name zu einer akzeptablen PDFreactor-API-Funktion?
    """
    if name in API_WHITELIST:
        return True
    if not name[3:]:
        return False
    if (name.startswith('set')
        or name.startswith('add')
        ):
        return True
    return False

COMMAND_TYPES = list(range(100, 103))
CT_CONTROL, CT_ASSIGNMENT, CT_API = COMMAND_TYPES
"""\
toc (h2, h3, h4) afterbegin of body
appendix afterbegin of "#appendix" (
  images grouped force,
  media grouped auto,
  tables grouped auto,
  literature sorted,
  standards sorted
)"""

CONTROL_COMMANDS = ('strict',
                    'toc',
                    'appendix',
                    'with_images',
                    )
SYMBOLS_MAP = {}
_tmp = [('True',  'on', 'yes'),
        ('False', 'off', 'no'),
        ('None',  'null', 'nothing', 'nil'),
        ]
for _tup in _tmp:
    first = _tup[0]
    for name in _tup:
        SYMBOLS_MAP[name.lower()] = first

def supported_assignment(stmt):
    """
    Take an Assignment (to some config key)
    and return a (key, value) tuple.

    >>> txt1 = 'config.disableLinks = True'
    >>> seq1 = list(gen_restricted_lines(txt1))
    >>> stmt1 = Assignment(seq1[0])
    >>> supported_statement(stmt1)
    ('disableLinks', True)
    """
    if stmt.name != 'config':
        raise ValueError('We support only assignments to config keys!'
                         " (%r != 'config')" % (
                         stmt.name,
                         ))
    pp(stmt.tokens)
    return stmt.key_and_value(skip=(None, '.'))


if __name__ == '__main__':
  if 1:
      txt1 = '''
      # don't forget that '.' OP:
      config.disableLinks = true
      '''
      stmts = list(gen_restricted_lines(txt1))
      stmt = stmts[0]
      print(stmt)
      kw = {'skip': (None, '.')}
      # Logging / Debugging:
      from pdb import set_trace
      set_trace()
      tup = stmt.key_and_value(**kw)
      print(tup)

      txt2 = '''
      # don't forget that '.' OP:
      config.outputFormat = {
          type: OutputType.JPEG,
      '''
  elif 0:
      txt = '''
      strict on
      # don't forget that '.' OP:
      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      # Standard library:
      from pprint import pprint

      # Logging / Debugging:
      from pdb import set_trace
      print(list(gen_restricted_lines(txt)))


      bogustxt = '''
      strict on
      # don't forget that '.' OP:
      config outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      print(list(gen_restricted_lines(bogustxt)))

      txt = '''
      strict on
      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      set_trace() # b 104
      for tokens in gen_restricted_lines(txt):
          pprint(tokens)
          # pprint(resolve_tokens(tokens))

  else:
    # Standard library:
    import doctest
    doctest.testmod()




"""
oldmethods: Map old API method calls (i.e., their arguments) to config keys of
the new PDFreactor client API,
and provide a conversion function.

As far as "symbols" are involved (like CLEANUP_CYBERNEKO and the like), we
have a mapping of the old names to the new values already in ./oldsymbols.py.
This is not used yet, though.

Here is what we have already.
We'll use a little test helper function:

    >>> from pdfreactor.parsecfg.convert import parse_configuration
    >>> def om(cfgtext):
    ...     config = parse_configuration(cfgtext, convert=convert_api_method)
    ...     return sorted(config.items())

    >>> om('enableDebugMode();setAddLinks(yes)')
    ... # doctest: +NORMALIZE_WHITESPACE
    [('debugSettings', {'appendLogs': True}),
     ('disableLinks',    False)]

NOTE:
    There has been a change in the symbols which have been provided for such
    calls; the CLEANUP_NONE symbol, to take one, was "renamed" to Cleanup.NONE
    (there is an 'Cleanup' attribute now to the PDFreactor class, which is a
    mini class itself and provides some names).
    The values have changed as well; while they used to be numeric, they are
    now strings, usually containing a trailing part of the name.

    There is no transformation of such values yet!
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from pdfreactor.parsecfg.convert import resolve_value

def negate(val):
    # negate the given value
    return not val

def unsupported(val):
    raise ValueError('The use of this method can\'t be converted '
                     'to some config key in the new PDFreactor API!')


map2 = {
    'setOutputFormat': [{
        'key': 'outputType',
        'subkey': 'type',
        }, {
        'key': 'outputType',
        'subkey': 'width',
        }, {
        'key': 'outputType',
        'subkey': 'height',
        }],
    'setCleanupTool': [{  # _new.Cleanup symbols
        'key': 'cleanupTool',
        }],
    'setEncoding': [{
        'key': 'encoding',
        }],
    'setJavaScriptMode': [{  # _new.JavaScriptMode symbols
        'key': 'javaScriptMode',
        }],
    'setAddBookmarks': [{
        'key': 'addBookmarks',
        }],
    'setAddLinks': [{
        'key': 'disableLinks',
        'convert': negate,
        }],
    'setAppendLog': [{
        'key': 'debugSettings',
        'subkey': 'appendLogs',
        }],
    'setDocumentType': [{
        'key': 'documentType',
        }],
    'setLicenseKey': [{  # not to be confused with the PDFreactor.apiKey attribute!
        'key': 'licenseKey',
        }],
    # https://www.pdfreactor.com/product/doc_html/#Logging
    'setLogLevel': [{  # _new.logLevel symbols
        'key': 'logLevel',
        }],
    }


def _specs(liz):
    if not liz:
        return
    elif not liz[1:]:
        yield 'val'
        return
    res = []
    for dic in liz:
        yield dic['subkey']


def convert_api_method(statement, config, control):  #  [[
    """
    Convert API methods (which are not used in the Python client API anymore)
    in config key assigments.

    Since this function is used as a `convert` argument to the
    parse_configuration function, we need to have a config and control dict,
    and we'll return a boolean value.

    >>> from pdfreactor.parsecfg.parse import generate_statements
    >>> def first_stmt(txt):
    ...     return list(generate_statements(txt))[0]
    >>> config = {}
    >>> ctrl = {}
    >>> stmt = first_stmt('the_answer=41')
    >>> convert_api_method(stmt, config, ctrl)
    False

    Well, this function doesn't convert assignments;
    so it returns False to allow you to try something else.

    For our further tests, we'll use a little test helper:
    >>> def cvt(txt, which='config'):
    ...     config = {}
    ...     control = {}
    ...     for stmt in generate_statements(txt):
    ...         res = convert_api_method(stmt, config, control)
    ...         print(res)
    ...     if which in ('config', 'both'):
    ...         print(sorted(config.items()))
    ...     if which in ('control', 'both'):
    ...         print(sorted(control.items()))

    >>> cvt('setAddLinks(yes)')
    True
    [('disableLinks', False)]

    >>> cvt('enableDebugMode()')
    True
    [('debugSettings', {'appendLogs': True})]

    """
    if not statement.is_method_call:
        return False

    grp0 = statement[0]
    the_name = grp0.dotted_name

    # TODO: create something generic, based on the map2 data!
    if the_name == 'setAddLinks':
        args = statement.method_args
        if args[1:]:
            tail = args[1]
            raise ValueError('Superfluous argument(s) in %(statement)r: '
                    '%(tail)r(...)' % locals())
        elif not args:
            val = True
        else:
            val = resolve_value(args[0])
            if not isinstance(val, int):
                raise ValueError('%(the_name)s(val) is expected to be a boolean '
                        '(or int); found %r' % (type(val),))
        control['add_links'] = val  # add_links_value
        config['disableLinks'] = not val
        return True
    elif the_name == 'enableDebugMode':
        args = statement.method_args
        if args:
            tail = args[0]
            raise ValueError('The %(the_name)s method doesn\'t expect arguments;'
                    'found %(tail)r' % locals())
        dic = config.setdefault('debugSettings', {})
        dic['appendLogs'] = True
        return True
    else:
        return False

if __name__ == '__main__':
  if 0:
    from six.moves import zip

    dictnames = set()
    nodicts = set()

    for name in sorted(map2.keys()):
        args = map2[name]
        specs = list(_specs(args))
        arglist = ', '.join(specs)
        print('%(name)s(%(arglist)s) -->' % locals())
        for spec, dic in zip(specs, args):
            try:
                key = dic['key']
            except KeyError:
                print('    (UNSUPPORTED)')
            else:
                if dic.get('convert'):
                    spec = '(converted) ' + spec
                if 'subkey' in dic:
                    subkey = dic['subkey']
                    assert key not in nodicts
                    dictnames.add(key)
                    print('    config[%(key)r][%(subkey)r] = %(spec)s' % locals())
                else:
                    assert key not in dictnames
                    nodicts.add(key)
                    print('    config[%(key)r] = %(spec)s' % locals())
  else:
    # Standard library:
    import doctest
    doctest.testmod()



"""
oldymbols: Provide the old symbols (old Python API, for PDFreactor 7),
but pointing to the new values.

For migration / conversion purposes only;
our export profiles are still based on the old API.

We are *not* interested in

- the actual (usually numeric) values of those old symbols
- inventing old-looking symbols for new functionality.

For the transformation of old API method calls to config variable assignments,
see ./oldmethods.py
"""

# Python compatibility:
from __future__ import absolute_import, print_function

# from pdfreactor.api import PDFreactor as _new

OLD2NEW = {

    # Cleanup: perform a cleanup when loading a document ...[
    'CLEANUP_NONE':	'Cleanup.NONE',	# no cleanup
    'CLEANUP_JTIDY':	'Cleanup.JTIDY',	# use JTidy
    'CLEANUP_CYBERNEKO':	'Cleanup.CYBERNEKO',	# use the CyberNeko HTML parser
    'CLEANUP_TAGSOUP':	'Cleanup.TAGSOUP',	# use the TagSoup HTML parser
    'CLEANUP_DEFAULT':	'Cleanup.CYBERNEKO',

    # ] Document type constants ... [

    'DOCTYPE_AUTODETECT':	'Doctype.AUTODETECT',
    'DOCTYPE_XML':	'Doctype.XML',
    'DOCTYPE_XHTML':	'Doctype.XHTML',
    'DOCTYPE_HTML5':	'Doctype.HTML5',
    'DOCTYPE_DEFAULT':	'Doctype.AUTODETECT',

    # ] Document default language constant ... [
    # NO SUCH SYMBOL in the new Python API, right?!
    'DOCUMENT_DEFAULT_LANGUAGE_DEFAULT': None,

    # ] Encryption constants ... [
    'ENCRYPTION_NONE':	'Encryption.NONE',
    'ENCRYPTION_40':	'Encryption.TYPE_40',  # 40 bit encryption
    'ENCRYPTION_128':	'Encryption.TYPE_128', # 128 bit encryption
    'ENCRYPTION_DEFAULT':	'Encryption.NONE',

    # ] Log level constants ... [
    # (old API method: setLogLevel(int)
    'LOG_LEVEL_NONE':	'LogLevel.NONE',
    'LOG_LEVEL_FATAL':	'LogLevel.FATAL',
    'LOG_LEVEL_WARN':	'LogLevel.WARN',
    'LOG_LEVEL_INFO':	'LogLevel.INFO',
    'LOG_LEVEL_DEBUG':	'LogLevel.DEBUG',
    'LOG_LEVEL_PERFORMANCE':	'LogLevel.PERFORMANCE',
    'LOG_LEVEL_DEFAULT':	'LogLevel.NONE',

    # ] Viewerpreferences ... [
    # (old API method: setViewerPreferences(int)
    'VIEWER_PREFERENCES_PAGE_LAYOUT_SINGLE_PAGE':	'ViewerPreferences.PAGE_LAYOUT_SINGLE_PAGE',
    'VIEWER_PREFERENCES_PAGE_LAYOUT_ONE_COLUMN':	'ViewerPreferences.PAGE_LAYOUT_ONE_COLUMN',
    'VIEWER_PREFERENCES_PAGE_LAYOUT_TWO_COLUMN_LEFT':	'ViewerPreferences.PAGE_LAYOUT_TWO_COLUMN_LEFT',
    'VIEWER_PREFERENCES_PAGE_LAYOUT_TWO_COLUMN_RIGHT':	'ViewerPreferences.PAGE_LAYOUT_TWO_COLUMN_RIGHT',
    'VIEWER_PREFERENCES_PAGE_LAYOUT_TWO_PAGE_LEFT':	'ViewerPreferences.PAGE_LAYOUT_TWO_PAGE_LEFT',
    'VIEWER_PREFERENCES_PAGE_LAYOUT_TWO_PAGE_RIGHT':	'ViewerPreferences.PAGE_LAYOUT_TWO_PAGE_RIGHT',
    'VIEWER_PREFERENCES_PAGE_MODE_USE_NONE':	'ViewerPreferences.PAGE_MODE_USE_NONE',
    'VIEWER_PREFERENCES_PAGE_MODE_USE_OUTLINES':	'ViewerPreferences.PAGE_MODE_USE_OUTLINES',
    'VIEWER_PREFERENCES_PAGE_MODE_USE_THUMBS':	'ViewerPreferences.PAGE_MODE_USE_THUMBS',
    'VIEWER_PREFERENCES_PAGE_MODE_FULLSCREEN':	'ViewerPreferences.PAGE_MODE_FULLSCREEN',
    'VIEWER_PREFERENCES_PAGE_MODE_USE_OC':	'ViewerPreferences.PAGE_MODE_USE_OC',
    'VIEWER_PREFERENCES_PAGE_MODE_USE_ATTACHMENTS':	'ViewerPreferences.PAGE_MODE_USE_ATTACHMENTS',
    'VIEWER_PREFERENCES_HIDE_TOOLBAR':	'ViewerPreferences.HIDE_TOOLBAR',
    'VIEWER_PREFERENCES_HIDE_MENUBAR':	'ViewerPreferences.HIDE_MENUBAR',
    'VIEWER_PREFERENCES_HIDE_WINDOW_UI':	'ViewerPreferences.HIDE_WINDOW_UI',
    'VIEWER_PREFERENCES_FIT_WINDOW':	'ViewerPreferences.FIT_WINDOW',
    'VIEWER_PREFERENCES_CENTER_WINDOW':	'ViewerPreferences.CENTER_WINDOW',
    'VIEWER_PREFERENCES_DISPLAY_DOC_TITLE':	'ViewerPreferences.DISPLAY_DOC_TITLE',
    'VIEWER_PREFERENCES_NON_FULLSCREEN_PAGE_MODE_USE_NONE':	'ViewerPreferences.NON_FULLSCREEN_PAGE_MODE_USE_NONE',
    'VIEWER_PREFERENCES_NON_FULLSCREEN_PAGE_MODE_USE_OUTLINES':	'ViewerPreferences.NON_FULLSCREEN_PAGE_MODE_USE_OUTLINES',
    'VIEWER_PREFERENCES_NON_FULLSCREEN_PAGE_MODE_USE_THUMBS':	'ViewerPreferences.NON_FULLSCREEN_PAGE_MODE_USE_THUMBS',
    'VIEWER_PREFERENCES_NON_FULLSCREEN_PAGE_MODE_USE_OC':	'ViewerPreferences.NON_FULLSCREEN_PAGE_MODE_USE_OC',
    'VIEWER_PREFERENCES_DIRECTION_L2R':	'ViewerPreferences.DIRECTION_L2R',
    'VIEWER_PREFERENCES_DIRECTION_R2L':	'ViewerPreferences.DIRECTION_R2L',
    'VIEWER_PREFERENCES_PRINTSCALING_NONE':	'ViewerPreferences.PRINTSCALING_NONE',
    'VIEWER_PREFERENCES_PRINTSCALING_APPDEFAULT':	'ViewerPreferences.PRINTSCALING_APPDEFAULT',
    'VIEWER_PREFERENCES_DUPLEX_SIMPLEX':	'ViewerPreferences.DUPLEX_SIMPLEX',
    'VIEWER_PREFERENCES_DUPLEX_FLIP_SHORT_EDGE':	'ViewerPreferences.DUPLEX_FLIP_SHORT_EDGE',
    'VIEWER_PREFERENCES_DUPLEX_FLIP_LONG_EDGE':	'ViewerPreferences.DUPLEX_FLIP_LONG_EDGE',
    'VIEWER_PREFERENCES_PICKTRAYBYPDFSIZE_FALSE':	'ViewerPreferences.PICKTRAYBYPDFSIZE_FALSE',
    'VIEWER_PREFERENCES_PICKTRAYBYPDFSIZE_TRUE':	'ViewerPreferences.PICKTRAYBYPDFSIZE_TRUE',

    'COLOR_SPACE_RGB':	'ColorSpace.RGB',
    'COLOR_SPACE_CMYK':	'ColorSpace.CMYK',

    # ] Conformance constants ... [
    # (old API method: setConformance(int)
    'CONFORMANCE_PDF':	'Conformance.PDF',
    'CONFORMANCE_PDFA1A':	'Conformance.PDFA1A',
    'CONFORMANCE_PDFA3A':	'Conformance.PDFA3A',
    'CONFORMANCE_PDFA':	'Conformance.PDFA3A',
    'CONFORMANCE_DEFAULT':	'Conformance.PDF',

    # ] Processing Preferences Constants ... [

    'PROCESSING_PREFERENCES_SAVE_MEMORY_IMAGES':	'ProcessingPreferences.SAVE_MEMORY_IMAGES',

    # ] Exceeding Content Constants ... [
    # (old API method: setLogExceedingContent(int, int)
    'EXCEEDING_CONTENT_ANALYZE_NONE':	'ExceedingContentAnalyze.NONE',
    'EXCEEDING_CONTENT_ANALYZE_CONTENT':	'ExceedingContentAnalyze.CONTENT',
    'EXCEEDING_CONTENT_ANALYZE_CONTENT_AND_STATIC_BOXES':	'ExceedingContentAnalyze.CONTENT_AND_STATIC_BOXES',
    'EXCEEDING_CONTENT_ANALYZE_CONTENT_AND_BOXES':	'ExceedingContentAnalyze.CONTENT_AND_BOXES',
    'EXCEEDING_CONTENT_AGAINST_NONE':	'ExceedingContentAgainst.NONE',
    'EXCEEDING_CONTENT_AGAINST_PAGE_BORDERS':	'ExceedingContentAgainst.PAGE_BORDERS',
    'EXCEEDING_CONTENT_AGAINST_PAGE_CONTENT':	'ExceedingContentAgainst.PAGE_CONTENT',
    'EXCEEDING_CONTENT_AGAINST_PARENT':	'ExceedingContentAgainst.PARENT',

    # ] Overlay & Merge Constants ... [
    # (old API method: setMergeMode(int)
    'MERGE_MODE_APPEND':	'MergeMode.APPEND',
    'MERGE_MODE_PREPEND':	'MergeMode.PREPEND',
    'MERGE_MODE_OVERLAY':	'MergeMode.OVERLAY',
    'MERGE_MODE_OVERLAY_BELOW':	'MergeMode.OVERLAY_BELOW',
    # (old API method: setOverlayRepeat(int)
    'OVERLAY_REPEAT_NONE':	'OverlayRepeat.NONE',
    'OVERLAY_REPEAT_LAST_PAGE':	'OverlayRepeat.LAST_PAGE',
    'OVERLAY_REPEAT_ALL_PAGES':	'OverlayRepeat.ALL_PAGES',

    # ] Multipage & Order Constants ... [
    # (old API method: setPagesPerSheetProperties(int)
    'PAGES_PER_SHEET_DIRECTION_RIGHT_DOWN':	'PagesPerSheetDirection.RIGHT_DOWN',
    'PAGES_PER_SHEET_DIRECTION_LEFT_DOWN':	'PagesPerSheetDirection.LEFT_DOWN',
    'PAGES_PER_SHEET_DIRECTION_RIGHT_UP':	'PagesPerSheetDirection.RIGHT_UP',
    'PAGES_PER_SHEET_DIRECTION_LEFT_UP':	'PagesPerSheetDirection.LEFT_UP',
    'PAGES_PER_SHEET_DIRECTION_DOWN_RIGHT':	'PagesPerSheetDirection.DOWN_RIGHT',
    'PAGES_PER_SHEET_DIRECTION_DOWN_LEFT':	'PagesPerSheetDirection.DOWN_LEFT',
    'PAGES_PER_SHEET_DIRECTION_UP_RIGHT':	'PagesPerSheetDirection.UP_RIGHT',
    'PAGES_PER_SHEET_DIRECTION_UP_LEFT':	'PagesPerSheetDirection.UP_LEFT',
    # (old API method: setPageOrder(string)
    'PAGE_ORDER_REVERSE':	'PageOrder.REVERSE',
    'PAGE_ORDER_ODD':	'PageOrder.ODD',
    'PAGE_ORDER_EVEN':	'PageOrder.EVEN',
    'PAGE_ORDER_BOOKLET':	'PageOrder.BOOKLET',
    'PAGE_ORDER_BOOKLET_RTL':	'PageOrder.BOOKLET_RTL',

    # ] Signing mode Constants ... [
    'KEYSTORE_TYPE_PKCS12':	'KeystoreType.PKCS12',
    'KEYSTORE_TYPE_JKS':	'KeystoreType.JKS',
    'SIGNING_MODE_SELF_SIGNED':	'SigningMode.SELF_SIGNED',
    'SIGNING_MODE_VERISIGN_SIGNED':	'SigningMode.VERISIGN_SIGNED',
    'SIGNING_MODE_WINCER_SIGNED':	'SigningMode.WINCER_SIGNED',

    # ] JavaScript mode Constants ... [
    # (old API method: setJavaScriptMode(int)
    'JAVASCRIPT_MODE_DISABLED':	'JavaScriptMode.DISABLED',
    'JAVASCRIPT_MODE_ENABLED':	'JavaScriptMode.ENABLED',
    'JAVASCRIPT_MODE_ENABLED_NO_LAYOUT':	'JavaScriptMode.ENABLED_NO_LAYOUT',
    'JAVASCRIPT_MODE_ENABLED_REAL_TIME':	'JavaScriptMode.ENABLED_REAL_TIME',
    'JAVASCRIPT_MODE_ENABLED_TIME_LAPSE':	'JavaScriptMode.ENABLED_TIME_LAPSE',
    'JAVASCRIPT_MODE_DEFAULT':	'JavaScriptMode.DISABLED',

    # ] HTTPS mode Constants ... [
    # (old API method: setHttpsMode(int)
    'HTTPS_MODE_STRICT':	'HttpsMode.STRICT',
    'HTTPS_MODE_LENIENT':	'HttpsMode.LENIENT',
    'HTTPS_MODE_DEFAULT':	'HttpsMode.STRICT',

    # ] Output Format Constants ... [
    # (old API method: setOutputFormat(int, int, int)
    'OUTPUT_FORMAT_PDF':	'OutputType.PDF',
    'OUTPUT_FORMAT_JPEG':	'OutputType.JPEG',
    'OUTPUT_FORMAT_PNG':	'OutputType.PNG',
    'OUTPUT_FORMAT_PNG_TRANSPARENT':	'OutputType.PNG_TRANSPARENT',
    'OUTPUT_FORMAT_BMP':	'OutputType.BMP',
    'OUTPUT_FORMAT_GIF':	'OutputType.GIF',
    'OUTPUT_FORMAT_PNG_AI':	'OutputType.PNG_AI',
    'OUTPUT_FORMAT_PNG_TRANSPARENT_AI':	'OutputType.PNG_TRANSPARENT_AI',
    'OUTPUT_FORMAT_TIFF_LZW':	'OutputType.TIFF_LZW',
    'OUTPUT_FORMAT_TIFF_PACKBITS':	'OutputType.TIFF_PACKBITS',
    'OUTPUT_FORMAT_TIFF_UNCOMPRESSED':	'OutputType.TIFF_UNCOMPRESSED',
    'OUTPUT_FORMAT_TIFF_CCITT_1D':	'OutputType.TIFF_CCITT_1D',
    'OUTPUT_FORMAT_TIFF_CCITT_GROUP_3':	'OutputType.TIFF_CCITT_GROUP_3',
    'OUTPUT_FORMAT_TIFF_CCITT_GROUP_4':	'OutputType.TIFF_CCITT_GROUP_4',
    }


if __name__ == '__main__':
    # Standard library:
    from pprint import pprint
    # Local imports:
    from pdfreactor.parsecfg.symbols import SYMBOL_STRINGS as _new

    removed = set()
    found_new = set()
    for key, val in OLD2NEW.items():
        if val is None:
            removed.add(key)
            continue
        found_new.add(val)
        dic = {
            'Old symbol': key,
            'New symbol': val,
            'New value': _new[val],
        }
        pprint(dic)
    print('%d symbols have been removed:' % len(removed))
    print(sorted(removed))
    print('%d symbols have new names:' % len(found_new))
    print(sorted(found_new))
    new_only = set(_new) - found_new
    print('%d symbols have been added:' % len(new_only))
    print(sorted(new_only))




# -*- coding: utf-8
"""
pdfreactor.parsecfg._tokensgroup: TokensGroup class

(helper class for the .parse module)
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types
from six.moves import range, zip
from six.moves import StringIO

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Standard library:
from collections import deque
# from tokenize import untokenize  (customize, perhaps?)
from tokenize import (
    COMMENT,
    ENDMARKER,
    INDENT,
    NAME,
    NEWLINE,
    NL,
    OP,
    STRING,
    TokenError,
    generate_tokens,
    tok_name,
    )

# visaplan:
from visaplan.tools.minifuncs import makeBool
from visaplan.tools.minifuncs import translate_dummy as _
from visaplan.tools.sequences import sequence_slide

# Local imports:
from pdfreactor.parsecfg._args import extract_input_specs
from pdfreactor.parsecfg._tokeninfo import TokenInfo


class TokensGroup(list):
    """
    A class to contain a group of tokens, as generated by the
    generate_token_groups function.

    >>> txt = 'config.disableLinks = true;'
    >>> groups = [TokensGroup(grp)
    ...           for grp in generate_token_groups(txt)]
    >>> dn = groups[0]  # a dotted name
    >>> dn
    <NAME config.disableLinks>
    >>> dn.is_terminator
    False
    >>> dn.is_dotted_name
    True
    >>> dn.names_list
    ['config', 'disableLinks']

    >>> op = groups[1]
    >>> op
    <OP '='>
    >>> op.is_terminator
    False
    >>> op.is_dotted_name
    False

    >>> sy = groups[2]  # another dotted name we'll consider a symbol
    >>> sy
    <NAME true>
    >>> sy.is_terminator
    False
    >>> sy.is_dotted_name
    True
    >>> sy.names_list
    ['true']

    >>> tg = groups[3]
    >>> tg
    <OP ';'>
    >>> tg.is_terminator
    True
    >>> tg.is_dotted_name
    False
    >>> tg.names_list
    """

    @property
    def text(self):
        return ''.join([tup[1] for tup in self])

    @property
    def ttype(self):
        """
        The token type for our group is the type of the first token
        """
        return self[0][0]

    @property
    def tname(self):
        """
        The name of our "token type", as given by the tok_name dict
        """
        return tok_name[self.ttype]

    @property
    def opens_brace(self):
        return self[0].opens_brace

    @property
    def closes_brace(self):
        return self[0].closes_brace

    @property
    def expects_brace(self):
        """
        This will yield a KeyError if not .opens_brace!
        """
        return self[0].expects_brace

    def __repr__(self):
        ttype = self.ttype
        if self.is_terminator:
            txt = self.text.strip()
            if txt:
                return '<%s %r>' % (self.tname, txt)
            else:
                return '<%s>' % (self.tname,)
        elif ttype == OP:
            mask = '<%s %r>'
        else:
            mask = '<%s %s>'
        return mask % (self.tname, self.text)

    @property
    def is_terminator(self):
        """
        Would this group of tokens terminate a statement?
        """
        ttype = self.ttype
        if ttype == OP:
            length = len(self)
            if length > 1:
                raise ValueError('Invalid operator %(self)r: '
                     'too long (%(length)d tokens)!' % locals())
            return self.text == ';'
        elif ttype == NAME:
            return False
        else:
            text = self.text.strip()
            return not text

    @property
    def is_dotted_name(self):
        """
        We use dotted names ...
        - as assignment targets, e.g. config.disableLinks
        - as PDFreactor symbols, e.g. Doctype.AUTODETECT
        """
        if self.ttype == NAME:
            for i, tok in zip(list(range(1, len(self))), self[1:]):
                odd = i % 2
                if odd:
                    if tok[1] != '.':
                        raise ValueError('Invalid dotted name %r: '
                                'expected odd tokens to be dots; '
                                'found %r!' % (self, tok[1]))
                elif tok[0] != NAME:
                    tname = tok_name[tok[0]]
                    raise ValueError('Invalid dotted name %r: '
                            'expected even tokens to be names; '
                            'found a %s %r!'
                            % (self, tname, tok[1]))
            return True
        return False

    @property
    def names_list(self):
        """
        The list of names; useful to compile a symbol name
        or the target variable and key
        """
        if self.ttype != NAME:
            return None
        res = []
        for i, tok in zip(list(range(len(self))), self):
            odd = i % 2
            if odd:
                if tok[1] != '.':
                    raise ValueError('Invalid dotted name %r: '
                            'expected odd tokens to be dots; '
                            'found %r!' % (self, tok[1]))
            elif tok[0] != NAME:
                tname = tok_name[tok[0]]
                raise ValueError('Invalid dotted name %r: '
                        'expected even tokens to be names; '
                        'found a %s %r!'
                        % (self, tname, tok[1]))
            else:
                res.append(tok[1])
        return res

    @property
    def dotted_name(self):
        the_list = self.names_list
        if the_list is None:
            return None
        return '.'.join(the_list)


def generate_token_groups(*args, **kwargs):
    r"""
    Generate token groups which we are willing to interpret

    We use the Python parser to parse to parse the given text
    and group tokens together which form some interesting unit.

    We'll use two little helper classes which are implemented above:

    - TokenInfo(tuple), to take a tuple as generated by the
      tokenize.generate_tokens function
    - TokensGroup(list), to take a list of adjacent tuples
      which toether form some useful entity, e.g. a dotted name.

    Both classes should be usable just like their parent classes (tuple and
    list); but they provide some handy properties which should ease our task.

    >>> def lotg(txt):  # list of token groups
    ...     return list(generate_token_groups(txt))

    Let's suppose we have the following configuration text:

    >>> txt = 'config.disableLinks = true'
    >>> lotg(txt)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<NAME config.disableLinks>, <OP '='>, <NAME true>, <ENDMARKER>]

    >>> def lotg(txt):
    ...     return [tg for tg in generate_token_groups(txt)]

    >>> lotg(txt)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<NAME config.disableLinks>, <OP '='>, <NAME true>, <ENDMARKER>]

    At this stage, we are not picky about what we accept;
    we are prepared to accept old (PDFreactor v7-) API method names
    and proprietary control statements:

    >>> txt = '''# some comment
    ... strict on
    ... setAddLinks(False); # another comment
    ... with_images off
    ... '''
    >>> lotg(txt)                             # doctest: +NORMALIZE_WHITESPACE
    [<NAME strict>, <NAME on>,
     <NEWLINE>,
     <NAME setAddLinks>, <OP '('>, <NAME False>, <OP ')'>, <OP ';'>,
     <NAME with_images>, <NAME off>,
     <NEWLINE>]

    >>> txt2 = '''
    ... setAddLinks(False)
    ... # Enable bookmarks in the PDF document
    ... setAddBookmarks(True)
    ... setCleanupTool(CLEANUP_NONE)
    ... setEncoding('UTF-8')
    ... setJavaScriptMode(JAVASCRIPT_MODE_ENABLED_NO_LAYOUT)
    ... '''
    >>> lotg(txt2)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<NAME setAddLinks>,
         <OP '('>, <NAME False>, <OP ')'>,
         <NEWLINE>,
     <NAME setAddBookmarks>,
         <OP '('>, <NAME True>, <OP ')'>,
         <NEWLINE>,
     <NAME setCleanupTool>,
         <OP '('>, <NAME CLEANUP_NONE>, <OP ')'>,
         <NEWLINE>,
     <NAME setEncoding>,
         <OP '('>, <STRING 'UTF-8'>, <OP ')'>,
         <NEWLINE>,
     <NAME setJavaScriptMode>,
         <OP '('>, <NAME JAVASCRIPT_MODE_ENABLED_NO_LAYOUT>, <OP ')'>,
         <NEWLINE>]

    Of course, the conversion to the respective new API config assignments is
    not done here.

    We may put multiple statements in one line, separated by semicolons:

    >>> txt3 = '; ;setA();setB()'
    >>> lotg(txt3)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<NAME setA>,
         <OP '('>,
         <OP ')'>,
         <OP ';'>,
     <NAME setB>,
         <OP '('>,
         <OP ')'>,
         <ENDMARKER>]

    Comments work as expected, of course:

    >>> txt3c = 'setA();#setB()'
    >>> lotg(txt3c)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<NAME setA>,
         <OP '('>,
         <OP ')'>,
         <OP ';'>]

    We mainly expect to see assignments for the current PDFreactor API
    which we support in convenient fashion.

    Instead of something like:

      config['outputFormat'] = {
          'type': PDFreactor.OutputType.JPEG,
          'width': 640,
          }

    (which is what you'll get in the end,
    and which might be supported by a future release),
    you may specify:

    >>> txt4 = '''
    ... config.outputFormat = {
    ...     type: OutputType.JPEG,
    ...     width: 640,
    ... }
    ... '''
    >>> lotg(txt4)                            # doctest: +NORMALIZE_WHITESPACE
    [<NAME config.outputFormat>,
     <OP '='>,
     <OP '{'>,
         <NL>,
         <NAME type>,  <OP ':'>, <NAME OutputType.JPEG>,
         <OP ','>,
         <NL>,
         <NAME width>, <OP ':'>, <NUMBER 640>,
         <OP ','>,
         <NL>,
     <OP '}'>,
     <NEWLINE>]

    We are not strictly limited to using the core PDFreactor API.
    Here is an example from our integration which provides the server-side
    creation of appendixes and tables of contents.

    The Python parser yields a newline after every comment:
    >>> txt5 = '''appendix (
    ... #  glossary,
    ...   images,
    ... )'''
    >>> [TokenInfo(tup) for tup in (generate_tokens(StringIO(txt5).readline))]
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [(NAME appendix), (OP '('), (NL),
     (COMMENT #  glossary,), (NL),
     (NAME images), (OP ','), (NL), (OP ')'), (ENDMARKER)]

    For proper processing, we need to avoid generating two newlines in
    sequence; thus, we drop those newlines after comments:
    >>> lotg(txt5)                            # doctest: +NORMALIZE_WHITESPACE
    [<NAME appendix>, <OP '('>, <NL>,
     <NAME images>, <OP ','>, <NL>, <OP ')'>, <ENDMARKER>]

    This is how our server-side table-of-contents creation would be
    configured:
    >>> lotg(
    ... '''toc (".headline-level-2", ".headline-level-3",
    ... ".headline-level-4")
    ... ''')                                  # doctest: +NORMALIZE_WHITESPACE
    [<NAME toc>, <OP '('>,
     <STRING ".headline-level-2">, <OP ','>,
     <STRING ".headline-level-3">, <OP ','>, <NL>,
     <STRING ".headline-level-4">, <OP ')'>, <NEWLINE>]

    (See the generate_statements doctest for the resulting statement for better
    readability.)
    Remember, the actuell processing is not part of this package;
    but you get an idea of what you could do.

    >>> lotg(                                 # doctest: +NORMALIZE_WHITESPACE
    ... "integrationStyleSheets = ['++resource++pdfreactor.plone/export.css']")
    [<NAME integrationStyleSheets>, <OP '='>, <OP '['>,
     <STRING '++resource++pdfreactor.plone/export.css'>, <OP ']'>, <ENDMARKER>]

    """
    input_specs = extract_input_specs(args, kwargs)
    text = input_specs.get('text')
    if text is None:
        fo = input_specs['file']
        alltokens = generate_tokens(fo.readline)
    elif not text:
        return
    else:
        alltokens = generate_tokens(StringIO(text).readline)
    verbose = kwargs.pop('verbose', False)
    firstn = 5 if verbose else 2
    braces_stack = deque()
    buf = []
    prev_ttype = None
    yield_singleton = False
    yield_sofar = False
    drop_this = False
    follows_terminator = True
    for toktup in alltokens:
        ti = TokenInfo(toktup)
        ttype = ti.ttype
        if ti.is_terminator:
            drop_this = follows_terminator or prev_ttype == COMMENT
            follows_terminator = True
            yield_singleton = not drop_this
            yield_sofar = True
            if 0 and  ti.text == ';' and braces_stack:
                open_braces = ''.join(braces_stack)
                raise ValueError("We don't expect %(ti)r in braces!"
                                 ' (%(open_braces)r)'
                                 % locals())
        else:
            follows_terminator = False
            if ttype == OP:
                token = ti.text
                if ti.opens_brace:
                    braces_stack.append(ti)
                elif ti.closes_brace:
                    if not braces_stack:
                        raise ValueError('Found closing %(ti)r, '
                            'but the braces stack is empty!'
                            % locals())

                    last_brace = braces_stack.pop()
                    expected = last_brace.expects_brace
                    if ti.text != expected:
                        raise ValueError('Unexpected closing brace %(ti)r; '
                                'expected %(expected)r!'
                                % locals())
                assert token != ';', ('Semicolons should have been treated '
                    'as terminators (above)!')
                if token != '.':
                    yield_sofar = True
                    yield_singleton = True
                elif ti.opens_brace or ti.closes_brace:
                    pass
                elif not buf:
                    raise ValueError("Can't start with a %(ti)r!"
                                     % locals())
                elif token == '.':
                    if prev_ttype != NAME:
                        so_far = ''.join([tup[1] for tup in buf])
                        raise ValueError("Won't accept a %(ti)r "
                            'after %(so_far)r!'
                            % locals())
            elif ttype == NAME and prev_ttype == NAME:
                # for control statements, e.g. 'strict off';
                # not implemented here:
                yield_sofar = True
            elif ttype in (NL, COMMENT):  # insignificant newline, or comment
                drop_this = True
            elif not buf and ttype in (INDENT,
                    ):
                drop_this = True

        if yield_sofar:
            if buf:
                yield TokensGroup(buf)
                buf = []
            yield_sofar = False

        if drop_this:
            assert not yield_singleton
            drop_this = False
        else:
            buf.append(TokenInfo(toktup[:firstn]))

        if yield_singleton:
            yield TokensGroup(buf)
            buf = []
            yield_singleton = False

        prev_ttype = ttype

    if braces_stack:
        cnt = len(braces_stack)
        braces = ' '.join([ti.text for ti in braces_stack])
        raise ValueError('Found %(cnt)d unclosed brace(s)! (%(braces)r)'
            % locals())
    if buf:
        yield TokensGroup(buf)


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8
"""
pdfreactor.parsecfg._statement: Statement class and generate_statements function

(helper class and generator for the .parse module)
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types
from six.moves import range, zip

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Standard library:
from collections import deque
# from tokenize import untokenize  (customize, perhaps?)
from tokenize import (
    COMMENT,
    ENDMARKER,
    INDENT,
    NAME,
    NEWLINE,
    NL,
    OP,
    STRING,
    TokenError,
    generate_tokens,
    tok_name,
    )

# visaplan:
from visaplan.tools.minifuncs import makeBool
from visaplan.tools.minifuncs import translate_dummy as _
from visaplan.tools.sequences import sequence_slide

# Local imports:
from pdfreactor.parsecfg._args import extract_input_specs
from pdfreactor.parsecfg._tokeninfo import TokenInfo
from pdfreactor.parsecfg._tokensgroup import TokensGroup, generate_token_groups

# from .oldmethods import map2
# from .oldsymbols import *


class Statement(list):
    """
    A class to represent a configuration statement which we'll interpret
    when using a textual configation to create (mainly) a config dictionary
    as expected e.g. by the PDFreactor.convert API method.

    For our logic, a Statement is just a list of TokensGroups
    (which may resemble e.g. a dotted name -- the main reason why we don't
    simply use tokens -- or an operator), which in turn are lists of TokenInfo
    objects.

    Given the following text:

    >>> txt = 'config.disableLinks = true'
    >>> groups = list(generate_token_groups(txt))
    >>> groups
    [<NAME config.disableLinks>, <OP '='>, <NAME true>, <ENDMARKER>]

    ... we can create the following statement (omitting that end marker):
    >>> stmt = Statement(groups[:-1])
    >>> stmt
    <Statement (config.disableLinks = true)>
    >>> stmt.is_assignment
    True
    >>> stmt.is_method_call
    False

    In the old (PDFreactor 7-) Python API, we had method calls:

    >>> txt = 'setAddLinks(true)'
    >>> groups = list(generate_token_groups(txt))
    >>> stmt = Statement(groups[:-1])
    >>> stmt
    <Statement (setAddLinks(true))>
    >>> stmt.is_assignment
    False
    >>> stmt.is_method_call
    True
    >>> stmt.method_args
    [<NAME true>]
    >>> stmt.name
    'setAddLinks'

    >>> txt = 'enableDebugMode()'
    >>> groups = list(generate_token_groups(txt))
    >>> stmt = Statement(groups[:-1])
    >>> stmt
    <Statement (enableDebugMode())>
    >>> stmt.is_assignment
    False
    >>> stmt.is_method_call
    True
    >>> stmt.method_args
    []
    >>> stmt.name
    'enableDebugMode'

    """

    def __repr__(self):
        return '<Statement (%(self)s)>' % locals()

    @property
    def is_assignment(self):
        if not self[1:]:
            return False
        grp2 = self[1]
        return grp2.ttype == OP and grp2.text == '='

    @property
    def is_method_call(self):
        if not self[1:]:
            return False
        name, op = self[:2]
        return name.ttype == NAME and op.ttype == OP and op.text == '('

    @property
    def method_args(self):
        if not self.is_method_call:
            return None
        res = []
        idx = 2
        for tg in self[2:]:
            if idx % 2:  # odd value: comma or ')' expected
                if tg.ttype != OP or tg.text not in ',)':
                    raise ValueError('Comma or ) expexted; found %(tg)r' % locals())
                if tg.text == ')':
                    break
            elif tg.ttype == OP:
                if tg.text == ')':
                    break
                else:
                    raise ValueError('Value expected; found %(tg)r' % locals())
            else:
                res.append(tg)
            idx += 1
        return res

    @property
    def name(self):
        return self[0].dotted_name

    def __str__(self):
        liz = []
        ins, app = 0, 0
        for prev_tg, current_tg, next_tg in sequence_slide(self):
            current_ttype = current_tg.ttype
            current_text = current_tg.text
            if current_ttype == OP:
                if current_text == '=':
                    ins, app = (1, 1)
                elif current_text == ',':
                    if not next_tg.closes_brace:
                        ins, app = (0, 1)
                    elif next_tg.text == ')':
                        ins, app = (0, 0)
                    else:
                        continue
                elif current_tg.opens_brace:
                    ins, app = (0, 0)
                elif current_tg.closes_brace:
                    ins, app = (0, 0)
                else:
                    ins, app = (0, 1)
            elif (current_ttype == NAME
                  and next_tg is not None
                  and next_tg.ttype == NAME
                  ):
                ins, app = (0, 1)
            else:
                ins, app = (0, 0)
            if ins:
                liz.append(' ')
            liz.append(current_text)
            if app and next_tg is not None:
                liz.append(' ')
        return ''.join(liz)


def generate_statements(*args, **kwargs):
    """
    Parse an input text and generate statements.

    We use the generate_token_groups function internally;
    why not use the same input data for our doctests?

    >>> def lost(txt):
    ...     return list(generate_statements(txt))
    >>> txt = 'config.disableLinks = true'
    >>> stmts = lost(txt)
    >>> stmts
    [<Statement (config.disableLinks = true)>]
    >>> stmt = stmts[0]
    >>> stmt.is_assignment
    True

    At this stage, we are not picky about what we accept;
    we are prepared to accept old (PDFreactor v7-) API method names
    and proprietary control statements:

    >>> txt = '''# some comment
    ... strict on
    ... setAddLinks(False); # another comment
    ... with_images off
    ... '''
    >>> lost(txt)                             # doctest: +NORMALIZE_WHITESPACE
    [<Statement (strict on)>,
     <Statement (setAddLinks(False))>,
     <Statement (with_images off)>]

    >>> txt2 = '''
    ... setAddLinks(False)
    ... # Enable bookmarks in the PDF document
    ... setAddBookmarks(True)
    ... setCleanupTool(CLEANUP_NONE)
    ... setEncoding('UTF-8')
    ... setJavaScriptMode(JAVASCRIPT_MODE_ENABLED_NO_LAYOUT)
    ... '''
    >>> lost(txt2)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<Statement (setAddLinks(False))>,
     <Statement (setAddBookmarks(True))>,
     <Statement (setCleanupTool(CLEANUP_NONE))>,
     <Statement (setEncoding('UTF-8'))>,
     <Statement (setJavaScriptMode(JAVASCRIPT_MODE_ENABLED_NO_LAYOUT))>]

    Of course, the conversion to the respective new API config assignments is
    not done here, where we just do some educated parsing.

    We may put multiple statements in one line, separated by semicolons:

    >>> txt3 = '; ;setA();setB()'
    >>> lost(txt3)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<Statement (setA())>,
     <Statement (setB())>]

    Comments work as expected, of course:

    >>> txt3c = 'setA();#setB()'
    >>> lost(txt3c)
    ...                                       # doctest: +NORMALIZE_WHITESPACE
    [<Statement (setA())>]

    We mainly expect to see assignments for the current PDFreactor API
    which we support in convenient fashion.

    Instead of something like:

    >>> txt4py = '''
    ... config['outputFormat'] = {
    ...     'type': PDFreactor.OutputType.JPEG,
    ...     'width': 640,
    ...     }
    ... '''

    (which is what you'll get in the end,
    and which might be supported by a future release),
    you may specify:

    >>> txt4 = '''
    ... config.outputFormat = {
    ...     type: OutputType.JPEG,
    ...     width: 640,
    ... }
    ... '''
    >>> lost(txt4)                            # doctest: +NORMALIZE_WHITESPACE
    [<Statement (config.outputFormat = {type:  OutputType.JPEG,
                                        width: 640})>]

    or:
    >>> txt5 = '''
    ... config.outputFormat.type = OutputType.JPEG
    ... config.outputFormat.width = 640
    ... '''
    >>> lost(txt5)                            # doctest: +NORMALIZE_WHITESPACE
    [<Statement (config.outputFormat.type = OutputType.JPEG)>,
     <Statement (config.outputFormat.width = 640)>]

    We can parse the Python assignment, however:
    >>> stmts = lost(txt4py)
    >>> stmts                                 # doctest: +NORMALIZE_WHITESPACE
    [<Statement (config['outputFormat'] =
        {'type': PDFreactor.OutputType.JPEG,
         'width': 640})>]

    We just don't recognize it as an assigment yet:
    >>> stmts[0].is_assignment                # doctest: +NORMALIZE_WHITESPACE
    False

    Now for our example configuration for server-side creation of appendixes
    and tables of contents (not implemented here).
    This is what it could be configured:
    >>> lost(
    ... '''toc (".headline-level-2", ".headline-level-3",
    ... ".headline-level-4")
    ... appendix (
    ... #  glossary,  (commented out)
    ...   images,
    ...   formulas,
    ... )
    ... ''')                                  # doctest: +NORMALIZE_WHITESPACE
    [<Statement (toc(".headline-level-2",
                     ".headline-level-3",
                     ".headline-level-4"))>,
     <Statement (appendix(images,
                          formulas,))>]

    ... or even:
    >>> lost(
    ... '''toc (h2, h3, h4) afterbegin of body
    ... appendix afterbegin of "#appendix" (
    ...   images grouped force,
    ...   media grouped auto,
    ...   tables grouped auto,
    ...   literature sorted,
    ...   standards sorted)
    ... ''')                                  # doctest: +NORMALIZE_WHITESPACE
    [<Statement (toc(h2, h3, h4)afterbegin of body)>,
     <Statement (appendix
        afterbegin of"#appendix"(images     grouped force,
                                 media      grouped auto,
                                 tables     grouped auto,
                                 literature sorted,
                                 standards  sorted))>]

    Here we have still a little problem; we need currently still be prepared to
    ignore left-over closing brackets:
    >>> lost(
    ... ''' appendix afterbegin of "#appendix" (
    ...   images grouped force
    ... )''')                           # doctest: +NORMALIZE_WHITESPACE +SKIP
    [<Statement (appendix
        afterbegin of"#appendix"(images grouped force)>,
        <Statement ())>]

    A comma at the EOL before the closing brace helps here:
    >>> lost(
    ... ''' appendix afterbegin of "#appendix" (
    ...   images grouped force,
    ... )''')                                 # doctest: +NORMALIZE_WHITESPACE
    [<Statement (appendix afterbegin of"#appendix"(images grouped force,))>]

    Remember, the actual processing is not part of this package;
    but you get an idea of what you could do.
    """
    input_specs = extract_input_specs(args, kwargs)
    token_groups = generate_token_groups(**input_specs)
    buf = []
    for prev_grp, grp, next_grp in sequence_slide(token_groups):
        if grp.is_terminator:
            if not buf:
                continue
            elif prev_grp.opens_brace:
                pass
            elif prev_grp.ttype == OP and prev_grp.text == ',':
                pass
            else:
                yield Statement(buf)
                buf = []
            continue
        buf.append(grp)
    assert not buf, "We get always a trailing terminator, right?!"


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



"""
pdfreactor.parsecfg: Configuration parser for PDFreactor client integrations
"""

# Python compatibility:
from __future__ import absolute_import, print_function

# from ._parse import make_parser, ControlStatement, ApiCall

if __name__ == '__main__':
  if 0:
      txt = '''
      strict on
      # don't forget that '.' OP:
      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      # Standard library:
      from pprint import pprint

      # Logging / Debugging:
      from pdb import set_trace
      print(list(gen_restricted_lines(txt)))


      bogustxt = '''
      strict on
      # don't forget that '.' OP:
      config outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      print(list(gen_restricted_lines(bogustxt)))

      txt = '''
      strict on
      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      set_trace() # b 104
      for tokens in gen_restricted_lines(txt):
          pprint(tokens)
          # pprint(resolve_tokens(tokens))

  else:
    # Standard library:
    import doctest
    doctest.testmod()



"""
pdfreactor.parsecfg.defaults: default values
"""

# Python compatibility:
from __future__ import absolute_import

from pdfreactor.defaults import default_config as _cfg

default_config_text = """\
disableLinks = %(disableLinks)r
""" % _cfg



# -*- coding: utf-8
"""
pdfreactor.parsecfg.symbols: known symbols of the current API

Instead of inspecting the PDFreactor class at runtime, we put the data in some
convenient dict object(s).
"""

# Python compatibility:
from __future__ import absolute_import

__all__ = [
    'SYMBOL_STRINGS',
    # 'SYMBOL_TUPLES',  (removed)
    'SIMPLE_SYMBOLS',
    ]

# Standard library:
from inspect import isclass

from pdfreactor.api import PDFreactor

# we'll see which one we'll use ..
SYMBOL_STRINGS = {}


for clsname in dir(PDFreactor):
    subclass = getattr(PDFreactor, clsname)
    if isclass(subclass):
        for symname in dir(subclass):
            if symname.startswith('_'):
                continue
            symbol = getattr(subclass, symname)
            if callable(symbol):
                raise ValueError('found callable %(symbol)r in %(subclass)r!'
                                 % locals())
            tup = (clsname, symname)
            key = '.'.join(tup)
            SYMBOL_STRINGS[key] = symbol

SIMPLE_SYMBOLS = {}
for val, keys in [
        (True,  ('True',  'on', 'yes')),
        (False, ('False', 'off', 'no')),
        (None,  ('None',  'null', 'nothing', 'nil')),
        ]:
    for key in keys:
        SIMPLE_SYMBOLS[key.lower()] = val

if __name__ == '__main__':
    # Standard library:
    from pprint import pprint
    pprint(SYMBOL_STRINGS)
    # pprint(SIMPLE_SYMBOLS)



# -*- coding: utf-8
"""
pdfreactor.parsecfg._args: Options evaluation

(helper functions for the .parse module)
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"


def extract_input_specs(args, kw):
    """
    Little options extraction helper; returns a dict.

    We expect a `text` or a `file` named option;
    if neither one is given, the first positional option is tried
    (which must be a string).

    >>> eis = extract_input_specs
    >>> eis([], {'text': 'some text'})
    {'text': 'some text'}
    >>> eis([], {'file': '<some file>'})
    {'file': '<some file>'}
    >>> eis([], {'text': 'some text',
    ...          'file': '<some file>'})
    Traceback (most recent call last):
      ...
    TypeError: Specified both a text and a file!
    >>> eis(('positional text',), {})
    {'text': 'positional text'}
    >>> eis((), {})
    Traceback (most recent call last):
      ...
    TypeError: Please specify a text or a file!

    >>> eis((42,), {})
    Traceback (most recent call last):
      ...
    TypeError: As positional input, we expect text; found <type 'int'>!

    If you specify a string instead of a list or tuple as a positional
    argument, you'll get:
    >>> eis('abc', {})
    Traceback (most recent call last):
      ...
    TypeError: Superfluous unnamed option(s)! ['b', 'c']
    >>> eis(['abc'], 'other crap')
    Traceback (most recent call last):
      ...
    TypeError: dict expected; got <type 'str'>!

    """
    res = {}
    if not isinstance(kw, dict):
        raise TypeError('dict expected; got %s!'
                        % (type(kw),
                           ))
    pop = kw.pop
    val = pop('text', None)
    if val is not None:
        res['text'] = val
    val = pop('file', None)
    if val is not None:
        if res:
            raise TypeError('Specified both a text and a file!')
        res['file'] = val
    if res:
        if args:
            raise TypeError('Unsupported unnamed option(s)!')
        return res
    if not args:
        raise TypeError('Please specify a text or a file!')
    elif isinstance(args, tuple):
        args = list(args)
    if not isinstance(args, list):
        args = list(args)
    val = args.pop(0)
    if isinstance(val, six_string_types):
        res['text'] = val
    else:
        raise TypeError('As positional input, we expect text; found %s!' % (
                        type(val),
                        ))
    if args:
        raise TypeError('Superfluous unnamed option(s)! %r'
                        % (args[:3],
                        ))
    return res


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8
"""
pdfreactor.parsecfg.parse: Configuration parser for PDFreactor client integrations

This module publishes the parsing functionality, based on the Python standard
tokenize module;  this includes two generators and three helper classes.

For the actual conversion to a config dictionary, see the convert module.
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types
from six.moves import range, zip

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Standard library:
# from tokenize import untokenize  (customize, perhaps?)
from tokenize import (
    COMMENT,
    ENDMARKER,
    INDENT,
    NAME,
    NEWLINE,
    NL,
    OP,
    STRING,
    TokenError,
    generate_tokens,
    tok_name,
    )

# Local imports:
from pdfreactor.parsecfg._statement import Statement, generate_statements
from pdfreactor.parsecfg._tokeninfo import TokenInfo
from pdfreactor.parsecfg._tokensgroup import TokensGroup, generate_token_groups

# from .oldmethods import map2
# from .oldsymbols import *


__all__ = [
    'generate_statements',    # ./_statement.py
    # helper functions:
    'generate_token_groups',  # ./_tokensgroup.py
    # helper classes:
    'TokenInfo',              # ./_tokeninfo.py
    'TokensGroup',            # ./_tokensgroup.py
    'Statement',              # ./_statement.py
    ]



"""
@@pdfreactor-config: parsed global settings
"""

# Python compatibility:
from __future__ import absolute_import

# Setup tools:
import pkg_resources

try:
    pkg_resources.get_distribution('pdfreactor.plone')
except pkg_resources.DistributionNotFound:
    HAVE_PDFREACTORPLONE = 0
else:
    HAVE_PDFREACTORPLONE = 1

# Zope:
from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
from zope.interface import implements

from pdfreactor.api import PDFreactor

# Local imports:
from pdfreactor.parsecfg.plone.interfaces import (
    IGetSettings,
    IPdfReactorConversionSettings,
    )

if HAVE_PDFREACTORPLONE:
    from pdfreactor.plone.config import DefaultSettingsView as BaseSettingsView
else:
    from Products.Five.browser import BrowserView as BaseSettingsView

# Local imports:
from ..convert import parse_configuration


class SettingsView(BaseSettingsView):

    implements(IGetSettings)  # usually IGetPdfReactorConversionSettings from
                              # pdfreactor.plone.interfaces

    def __call__(self):
        """
        Get the PDFreactor config(uration) settings as configured in the Plone
        registry.  You might want to inject contextual configuration changes as
        well.
        """
        context = aq_inner(self.context)
        registry = getToolByName(context, 'portal_registry')
        dic = registry.forInterface(IPdfReactorConversionSettings)
        config_text = dic.config_text
        config = parse_configuration(config_text)
        return config



# -*- coding: utf-8
"""
pdfreactor.parsecfg._tokeninfo: TokenInfo class

(helper class for the .parse module)
"""

# Python compatibility:
from __future__ import absolute_import

from six.moves import StringIO

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Standard library:
# from tokenize import untokenize  (customize, perhaps?)
from tokenize import (
    COMMENT,
    ENDMARKER,
    INDENT,
    NAME,
    NEWLINE,
    NL,
    OP,
    STRING,
    TokenError,
    generate_tokens,
    tok_name,
    )


class TokenInfo(tuple):
    """
    A class to provide information about a token
    as represented by a tuple, generated by the tokenize.generate_tokens
    function.

    This is mainly for internal use by the generate_token_groups function.

    >>> txt = 'empty = "";'
    >>> toklist = list(generate_tokens(StringIO(txt).readline))
    >>> len(toklist)
    5
    >>> name, op, val, semic, term = toklist

    >>> name_info = TokenInfo(name)
    >>> name_info
    (NAME empty)
    >>> name_info.is_terminator
    False

    >>> op_info = TokenInfo(op)
    >>> op_info
    (OP '=')
    >>> op_info.is_terminator
    False

    >>> val_info = TokenInfo(val)
    >>> val_info
    (STRING "")
    >>> val_info.is_terminator
    False

    >>> semic_info = TokenInfo(semic)
    >>> semic_info
    (OP ';')
    >>> semic_info.is_terminator
    True

    >>> term_info = TokenInfo(term)
    >>> term_info
    (ENDMARKER)
    >>> term_info.is_terminator
    True

    >>> txt = 'disableLinks(true)'
    >>> toklist = list(generate_tokens(StringIO(txt).readline))
    >>> len(toklist)
    5
    >>> name, op1, val, op2, term = toklist

    >>> name_info = TokenInfo(name)
    >>> name_info
    (NAME disableLinks)
    >>> name_info.is_terminator
    False

    >>> op1_info = TokenInfo(op1)
    >>> op1_info
    (OP '(')
    >>> op1_info.is_terminator
    False
    >>> op1_info.opens_brace
    True
    >>> op1_info.expects_brace
    ')'

    >>> val_info = TokenInfo(val)
    >>> val_info
    (NAME true)
    >>> val_info.is_terminator
    False

    >>> op2_info = TokenInfo(op2)
    >>> op2_info
    (OP ')')
    >>> op2_info.is_terminator
    False
    >>> op2_info.opens_brace
    False
    >>> op2_info.closes_brace
    True
    >>> op2_info.expects_brace
    Traceback (most recent call last):
      ...
    KeyError: ')'

    >>> term_info = TokenInfo(term)
    >>> term_info
    (ENDMARKER)
    >>> term_info.is_terminator
    True

    """
    @property
    def text(self):
        return self[1]

    @property
    def ttype(self):
        """
        The token type for our group is the type of the first token
        """
        return self[0]

    @property
    def tname(self):
        """
        The name of our "token type", as given by the tok_name dict
        """
        return tok_name[self.ttype]

    def __repr__(self):
        ttype = self.ttype
        if self.is_terminator:
            txt = self.text.strip()
            if txt:
                return '(%s %r)' % (self.tname, txt)
            else:
                return '(%s)' % (self.tname,)
        elif ttype == OP:
            mask = '(%s %r)'
        else:
            mask = '(%s %s)'
        return mask % (self.tname, self.text)

    @property
    def is_terminator(self):
        """
        Would this group of tokens terminate a statement?
        """
        ttype = self.ttype
        if ttype == OP:
            return self.text == ';'
        elif ttype == NAME:
            return False
        else:
            text = self.text.strip()
            return not text

    @property
    def _braces_dict(self):
        return {
            '{': '}',
            '(': ')',
            '[': ']',
            }

    @property
    def opens_brace(self):
        if self.ttype != OP:
            return False
        return self.text in self._braces_dict

    @property
    def expects_brace(self):
        """
        This will yield a KeyError if not .opens_brace!
        """
        return self._braces_dict[self.text]

    @property
    def closes_brace(self):
        if self.ttype != OP:
            return False
        return self.text in list(self._braces_dict.values())


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: et ts=8 sw=4 sts=4 si tw=79 cc=+1
"""Installer for the pdfreactor.parsecfg package."""
from __future__ import absolute_import
from __future__ import print_function

from setuptools import find_packages
from setuptools import setup
from os.path import isfile

package_name = 'pdfreactor.parsecfg'

# -------------------------------------------- [ get the version ... [
def read_version(fn, sfn):
    main = open(fn).read().strip()
    if sfn is not None and isfile(sfn):
        suffix = valid_suffix(open(sfn).read().strip())
    else:
        suffix = ''
    return main + suffix
    # ... get the version ...
def valid_suffix(suffix):
    """
    Enforce our suffix convention
    """
    suffix = suffix.strip()
    if not suffix:
        return suffix
    allowed = set('edv.0123456789rcpost')
    disallowed = set(suffix).difference(allowed)
    if disallowed:
        disallowed = ''.join(sorted(disallowed))
        raise ValueError('Version suffix contains disallowed characters'
                         ' (%(disallowed)s)'
                         % locals())
    chunks = suffix.split('.')
    chunk = chunks.pop(0)
    if chunk and not chunk.startswith('rc') and not chunk.startswith('post'):
        raise ValueError('Version suffix must start with "."'
                         ' (%(suffix)r)'
                         % locals())
    if not chunks:
        raise ValueError('Version suffix is too short'
                         ' (%(suffix)r)'
                         % locals())
    for chunk in chunks:
        if not chunk:
            raise ValueError('Empty chunk %(chunk)r in '
                             'version suffix %(suffix)r'
                             % locals())
        char = chunk[0]
        if char in '0123456789':
            raise ValueError('Chunk %(chunk)r of version suffix %(suffix)r'
                             ' starts with a digit'
                             % locals())
        char = chunk[-1]
        if char not in '0123456789':
            raise ValueError('Chunk %(chunk)r of version suffix %(suffix)r'
                             ' doesn\'t end with a digit'
                             % locals())
    return suffix  # ... valid_suffix
    # ... get the version ...
    # ... get the version ...
VERSION = read_version('VERSION',
                       'VERSION_SUFFIX')
# -------------------------------------------- ] ... get the version ]


# ------------------------------------------- [ for setup_kwargs ... [
long_description = '\n\n'.join([
    open('README.rst').read(),
    open('CONTRIBUTORS.rst').read(),
    open('CHANGES.rst').read(),
])

# see as well --> src/pdfreactor/parsecfg/configure.zcml:
namespace = 'pdfreactor'
packages = find_packages('src', exclude=['visaplan', 'visaplan.*'])

def github_urls(package, **kwargs):
    pop = kwargs.pop
    pkg_list = package.split('.')
    res = {}
    readthedocs = pop('readthedocs', False)
    if readthedocs:
        if readthedocs in (1, True):
            readthedocs = ''.join(pkg_list)
        res['Documentation'] = \
            'https://%(readthedocs)s.readthedocs.io' % locals()
        assert 'docs' not in kwargs
    else:
        docs = pop('docs', None)
        if docs is None:
            res['Documentation'] = 'https://pypi.org/project/%(package)s' \
                                   % locals()
        elif docs:
            res['Documentation'] = docs
    if not pop('github', 1):
        assert not kwargs
        return res
    pop_user = pop('pop_user', False)
    if pop_user:
        assert 'pick_user' not in kwargs
        assert 'user' not in kwargs
        user = pkg_list.pop(0)
        package = '.'.join(pkg_list)
    else:
        pick_user = pop('pick_user', 'user' not in kwargs)
        given_user = pop('user', None)
        if pick_user:
            user = pkg_list[0]
            if given_user is not None and given_user != user:
                raise ValueError('given user %(given_user)r mismatches '
                                 'user picked from package %(user)r!'
                                 % locals())
        elif given_user is not None:
            user = given_user
        else:
            raise ValueError('no user given nor picked!')
    if pop('travis', False):  # reqires github to be trueish
        res.update({  # CHECKME: is there a de-facto standard key for this?
            'Tests': 'https://travis-ci.org/%(user)s/%(package)s' % locals()
            })
    base = 'https://github.com/%(user)s/%(package)s' % locals()
    res.update({
        'Source': base,
        'Tracker': base + '/issues',
        })
    return res
project_urls = github_urls(package_name,
                           user='visaplan')
# ------------------------------------------- ] ... for setup_kwargs ]

setup_kwargs = dict(
    name=package_name,
    version=VERSION,
    description="PDFreactor configuration parser",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    # Get more from https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 4.3",
        'Framework :: Zope2',
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
    # keywords='Python Plone',
    author='Tobias Herp',
    author_email='tobias.herp@visaplan.com',
    project_urls=project_urls,
    license='MIT License',
    packages=packages,
    namespace_packages=[namespace],
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'pdfreactor-api',
        'six',
        'visaplan.tools',  # sequence_slide
    ],
    extras_require={
        'plone': [
            'pdfreactor.plone',
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
if 0:
    from pprint import pprint
    del setup_kwargs['long_description']
    pprint(setup_kwargs)
    raise SystemExit(1)
setup(**setup_kwargs)



# -*- coding: utf-8
"""
pdfreactor.parsecfg.convert: convert textual configuration to config dict
"""

# Python compatibility:
from __future__ import absolute_import

__all__ = [
    'parse_configuration',  # (**named options)
    # Toolset for own conversions:
    'update_config_dict',   # ...(config, statement)
    'resolve_value',        # ...(TokenGroup [,factory])
    ]

# Standard library:
from tokenize import (
    NAME,
    NUMBER,
    OP,
    STRING,
    )

# We might get along without this, but we'd like to have more tests first:
from visaplan.tools.sequences import sequence_slide

# Local imports:
from pdfreactor.parsecfg._args import extract_input_specs
from pdfreactor.parsecfg._configkeys import _factories
from pdfreactor.parsecfg.parse import generate_statements
from pdfreactor.parsecfg.symbols import SYMBOL_STRINGS, SIMPLE_SYMBOLS

# ... for doctests:
from pdfreactor.parsecfg._tokensgroup import TokensGroup
from pdfreactor.parsecfg._tokensgroup import generate_token_groups


def parse_configuration(*args, **kwargs):
    r"""
    Parse a textual configuration specification and return or modify a config
    dictionary (to be thrown e.g. at the PDFreactor.convert API method).

    >>> txt='config.outputFormat.type = OutputType.JPEG'

    If only one unnamed option is given, it is expected to be text, and the
    resulting config dict is returned:

    >>> config = parse_configuration(txt)
    >>> config
    {'outputFormat': {'type': 'JPEG'}}

    The configuration text can be specified as a named `text` option as well,
    or alternatively (UNTESTED) as a readable file object.

    An empty configuration text yields an empty dict, of course:
    >>> parse_configuration('')
    {}

    We won't be fooled by illegal options, of course:
    >>> parse_configuration('', no_unused=False)
    Traceback (most recent call last):
      ...
    TypeError: Found unsupported option(s)! ('no_unused')

    You might have a default configuration and want to apply changes to it.
    In this case you'd specify a named option `config` which will then be
    updated:

    >>> parse_configuration('config.outputFormat.width = 640', config=config)
    >>> config                                # doctest: +NORMALIZE_WHITESPACE
    {'outputFormat': {'width': 640,
                      'type': 'JPEG'}}

    >>> parse_configuration('''config.outputFormat = {
    ... width: 123,
    ... type: 'PNG',
    ... }''')
    {'outputFormat': {'width': 123, 'type': 'PNG'}}
    >>> parse_configuration('''outputFormat = {
    ... width: 234,
    ... type: 'JPEG',
    ... }''')
    {'outputFormat': {'width': 234, 'type': 'JPEG'}}
    >>> parse_configuration('''config.someList = (1, 2,3)''')
    {'someList': (1, 2, 3)}

    You might have special commands in your configuration which can't be taken
    as config assignments.  By default, these will cause the conversion to
    fail:
    >>> fancy_config = '''strict on
    ... config.outputFormat.type = OutputType.PNG_TRANSPARENT
    ... '''
    >>> parse_configuration(fancy_config)
    Traceback (most recent call last):
      ...
    ValueError: Found 1 unsupported statement(s) (strict on)

    You may specify an `unused` list to handle this:
    >>> unused = []
    >>> parse_configuration(fancy_config, unused=unused)
    {'outputFormat': {'type': 'PNG_TRANSPARENT'}}
    >>> unused
    [<Statement (strict on)>]

    Alternatively you might specify an own conversion function (named option
    `convert`) which is tried first for each statement and handles those, as
    we call them, "control commands".  If it returns something falsy, the
    standard processing will be tried next:

    >>> parse_configuration('', convert=42)   # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    ValueError: 'convert' option must be some function
                  bool <= f(statement, config, control);
                found <type 'int'>!

    Since the license key (unless installed on the server; see below)
    needs to be injected in the config dict,
    you might configure it here as well:

    >>> parse_configuration(""
    ... "licenseKey = '<license>... some XML text ...</license>'")
    {'licenseKey': '<license>... some XML text ...</license>'}

    But please note that we currently don't support string expressions,
    so you'd really need to but this in one long line:

    >>> parse_configuration('''
    ... licenseKey = ('<license>... some XML text ...'
    ...               '... implicitly concatened by Python ...</license>')
    ... ''')                                  # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    ValueError: Found 1 unsupported statement(s)
        ('... implicitly concatened by Python ...</license>'))
    >>> parse_configuration('''
    ... licenseKey = '<license>... some XML text ...' \
    ...            + '... explicitly concatened by Python ...</license>'
    ... ''')                                  # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    ValueError: <Statement (licenseKey = '<license>... some XML text ...'+
        '... explicitly concatened by Python ...</license>')>:
        nothing expected after <STRING '<license>... some XML text ...'>;
        2 surplus token( group)s found!

    In fact, we consider the licenseKey to be connection-related; e.g. when
    using our Plone integration, it can be configured in the
    IPdfReactorConnectionSettings interface.

    BTW, we support it this way because the PDFreactor API supports it.
    It is considered much better, however, to install the license key on the
    PDFreactor server, preventing it from being sent around all the time;
    to restrict access to your service, you have the options

    - to use API keys (see our pdfreactor.apikey package),
      or
    - to restrict access by IP address filtering.

    For some configuration values we need to take the model into account;
    e.g., stylesheets are not specified directly (as URLs only)
    but need to be given as a dict:

    >>> parse_configuration('''
    ... integrationStyleSheets = ['++resource++pdfreactor.plone/export.css']
    ... ''')                                  # doctest: +NORMALIZE_WHITESPACE
    {'integrationStyleSheets':
        [{'uri': '++resource++pdfreactor.plone/export.css'}]}

    """
    input_specs = extract_input_specs(args, kwargs)
    inspect_parse_specs(kwargs)  # all are keyword-only

    config = kwargs['config']
    control = kwargs['control']
    unused = kwargs['unused']
    convert = kwargs['convert']
    dict_style = kwargs['style'] == 'dict'
    statements = generate_statements(**input_specs)
    for stmt in statements:
        if convert is not None and convert(stmt, config, control):
            continue
        elif not stmt.is_assignment:
            unused.append(stmt)
        elif dict_style:
            update_config_dict(config, stmt)
        else:
            amend_config_list(config, stmt)

    if unused and kwargs['no_unused']:
        cnt = len(unused)
        first = unused[0]
        more = cnt > 1 and ' (...)' or ''
        raise ValueError('Found %(cnt)d unsupported statement(s) '
                         '(%(first)s%(more)s)'
                         % locals())
    if kwargs['return_configuration']:
        return config


def inspect_parse_specs(kw):
    """
    Helper: Evaluate named options for the parse_configuration function above

    >>> kw = {}
    >>> inspect_parse_specs(kw)
    >>> sorted(kw.items())                    # doctest: +NORMALIZE_WHITESPACE
    [('config', {}),
     ('control', {}),
     ('convert', None),
     ('no_control', True),
     ('no_unused', True),
     ('return_configuration', True),
     ('style', 'dict'),
     ('unused', [])]

    If a config dict is given, it is not returned but modified in-place:
    >>> config = {'some': 'configuration'}
    >>> kw = {'config': config}
    >>> inspect_parse_specs(kw)
    >>> kw['return_configuration']
    False

    Some day, perhaps, the given configuration may be alternatively a list:
    >>> config = [('some', 'configuration')]
    >>> kw = {'config': config}
    >>> inspect_parse_specs(kw)                               # doctest: +SKIP
    >>> sorted(kw.items())              # doctest: +NORMALIZE_WHITESPACE +SKIP
    [('config', [('some', 'configuration')]),
     ('control', {}),
     ('convert', None),
     ('no_control', True),
     ('no_unused', True),
     ('return_configuration', False),
     ('style', 'list'),
     ('unused', [])]

    This is not implemented (yet), however; so we'll get an error:
    >>> inspect_parse_specs(kw)
    Traceback (most recent call last):
      ...
    ValueError: config structure 'list' style is not (yet?) supported
    """
    pop = kw.pop
    no_config = 'config' not in kw
    upd = {'return_configuration': no_config}
    if no_config:
        fact = pop('config_factory', dict)
        kw['config'] = fact()
    val = kw['config']  # will be used
    if isinstance(val, dict):
        upd['style'] = 'dict'
    elif isinstance(val, list):
        upd['style'] = 'list'
        raise ValueError('config structure %(style)r style is not (yet?)'
                         ' supported' % upd)
    else:
        raise ValueError('config structure must be a list or dict; found %s!'
                         % (type(val),
                            ))

    # control: a namespace for 'strict' (...) specifications
    #          (to be implemented in your own application)
    upd['no_control'] = no_control = 'control' not in kw
    if no_control:
        kw['control'] = {}
    elif not isinstance(kw['control'], dict):
        raise ValueError('control structure must be some dict; found %s!'
                         % (type(kw['control']),
                            ))

    # unused: a list of unused statements.
    #
    # You might have statements which don't result in some config assignment,
    # and which are meant to be used by your application code (typically before
    # requesting the PDF conversion).
    # Specify a list to handle them yourself after parsing;
    # if you don't, we'll consider such statements an error.
    upd['no_unused'] = no_unused = 'unused' not in kw
    if no_unused:
        kw['unused'] = []
    elif not isinstance(kw['unused'], list):
        raise ValueError("'unused' option must be some list; found %s!"
                         % (type(kw['unused']),
                            ))

    # convert: a function for alternate conversions
    #
    # This is the plugin facility for custom conversions.
    # You might specify a function with the following signature:
    #
    #   bool <= f(stmt, config, control)
    #
    # The function should inspect the statement <stmt> (just a list of groups
    # of tokens; see our Statement class) and apply changes to the config
    # structure and / or control dict.
    # If it returns True, the statement is considered dealt with;
    # otherwise we'll check whether we recognize it as an assignment.
    # If we don't, it is put to the 'unused' list.
    val = kw.setdefault('convert', None)
    assert 'convert' in kw
    if val is None:
        pass
    elif not callable(val):
        raise ValueError("'convert' option must be some function"
                         ' bool <= f(statement, config, control);'
                         ' found %s!'
                         % (type(kw['convert']),
                            ))
    valid_options = set(['config', 'control',
                         'convert',
                         'unused'])
    invalid = set(kw) - valid_options
    if invalid:
        raise TypeError('Found unsupported option(s)! (%r)'
                        % (sorted(invalid)[0],
                           ))
    kw.update(upd)


def checked_reactor_dest(tg):
    """
    Take a TokenGroup and return a dict to tell how to deal with it.

    We'll use a handy little test helper here:

    >>> def f(txt):
    ...     groups = list(generate_token_groups(txt))
    ...     dest = groups[0]
    ...     return sorted(checked_reactor_dest(dest).items())

    >>> f('config.disableLinks')              # doctest: +NORMALIZE_WHITESPACE
    [('key', 'disableLinks'),
     ('key_factory', <type 'bool'>),
     ('subkey', None),
     ('tip_factory', <type 'bool'>),
     ('tip_keys', ['disableLinks'])]

    In fact the leading 'config.' is optional:
    >>> f('disableLinks')                     # doctest: +NORMALIZE_WHITESPACE
    [('key', 'disableLinks'),
     ('key_factory', <type 'bool'>),
     ('subkey', None),
     ('tip_factory', <type 'bool'>),
     ('tip_keys', ['disableLinks'])]

    The most interesting case are dotted assignment targets:
    >>> f('outputFormat.type')                # doctest: +NORMALIZE_WHITESPACE
    [('key',            'outputFormat'),
     ('key_factory',    <type 'dict'>),
     ('subkey',         'type'),
     ('subkey_factory', <type 'str'>),
     ('tip_factory', <type 'str'>),
     ('tip_keys', ['outputFormat', 'type'])]

    A bare 'config' won't be enough, though:
    >>> f('config')                           # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    ValueError: Invalid assigment destination <NAME config>:
        no names left behind 'config'!
    """
    names = list(tg.names_list)
    if names[0] == 'config':
        # currently we don't have a complete list of possible names and their types yet ...
        # so we just check that there is more:
        head = names.pop(0)
        if not names:
            raise ValueError('Invalid assigment destination %(tg)r: '
                    'no names left behind %(head)r!'
                    % locals())
    key = names.pop(0)
    tip_keys = [key]
    res = {
        'key': key,
        }
    key_factory = _factories.get((key,), None)

    if names:
        subkey = names.pop(0)
        if names:
            raise ValueError('Invalid assigment destination %(tg)r: '
                    'too many levels!'
                    % locals())
        tip_keys.append(subkey)
        subkey_factory = _factories.get((key, subkey), None)
        if key_factory is None:
            # we have subkeys, so this should be a dict:
            key_factory = dict
        else:
            assert key_factory is dict
        res.update({
            'subkey': subkey,
            'subkey_factory': subkey_factory,
            'tip_factory': subkey_factory,
            })
    else:
        res.update({
            'tip_factory': key_factory,
            'subkey': None,
            })
    res.update({
        'key_factory': key_factory,
        'tip_keys': tip_keys,
        })
    return res


_more_factories = {}
def make_af():
    from os.path import dirname, join, sep
    mydir = dirname(__file__)
    devmode = 'src' in mydir.split(sep)
    modfile = join(mydir, '_configkeys.py')

    def add_new_factory(fact, *keys):
        if isinstance(keys, list):
            keys = tuple(keys)
        if keys in _more_factories:
            oldfact = _more_factories[keys]
            if oldfact is fact:
                return 0
            raise ValueError('Factories conflict for %(keys)r: '
                    '%(fact)r mismatches %(oldfact)r!'
                    % locals())
        else:
            assert callable(fact), ('Something callable expected;'
                    ' %(fact)r found!'
                    ) % locals()
            _more_factories[keys] = fact
            return 1

    def amend_module(fact, *keys):
        if add_new_factory(fact, *keys):
            if isinstance(keys, list):
                keys = tuple(keys)
            with open(modfile, 'a') as mf:
                txt = '_factories[%(keys)s] = %(fact)s\n' % locals()
                mf.write(txt)

    if devmode and 0:
        return amend_module
    return add_new_factory
another_factory = make_af()


def set_value(dic, key, tg, fact=None):
    """
    Helper function for update_config_dict
    """
    assert isinstance(dic, dict)
    assert isinstance(key, basestring)
    assert isinstance(tg, TokensGroup)
    dic[key] = resolve_value(tg, fact)


def resolve_value(tg, fact=None):
    """
    Resolve the value of a TokenGroup to some Python value.

    Let's use a little test helper:

    >>> def f(txt, factory=None):
    ...     groups = list(generate_token_groups(txt))
    ...     group = groups[0]
    ...     return resolve_value(group, factory)

    We support PDFreactor symbols:

    >>> f('OutputType.JPEG')
    'JPEG'

    ... unless mistyped:
    >>> f('outputType.JPEG')
    Traceback (most recent call last):
      ...
    ValueError: <NAME outputType.JPEG>: No known symbol!

    We support simple (case insensitive) aliases for boolean and None values:

    >>> f('OFF')
    False
    >>> f('Yes')
    True
    >>> f('NIL') is None
    True
    >>> f('nothing') is None
    True

    We support strings:
    >>> f('"Some string"')
    'Some string'

    We support numbers:
    >>> f('42')
    42
    >>> f('3.141')
    3.141

    You may specify a custom factory function:
    >>> def foolme(s):
    ...     return 3.141 if '.' in s else 42
    >>> f('1.2', foolme)
    3.141
    >>> f('1', foolme)
    42

    """
    if fact is not None:
        assert callable(fact)
    if tg.is_dotted_name:
        txt = tg.text
        if txt in SYMBOL_STRINGS:
            val = SYMBOL_STRINGS[txt]
        else:
            ltxt = txt.lower()
            if ltxt in SIMPLE_SYMBOLS:
                val = SIMPLE_SYMBOLS[ltxt]
            else:
                raise ValueError('%(tg)s: No known symbol!'
                        % locals())
    elif tg.ttype == NUMBER:
        txt = tg.text
        if fact is not None:
            val = fact(txt)
        elif '.' in txt:
            val = float(txt)
        else:
            val = int(txt)
    elif tg.ttype == STRING:
        val = eval(tg.text)
        if fact is not None:
            val = fact(val)
    else:
        raise ValueError('Don\'t know how to use the value %(tg)s!'
                         % locals())
    return val


def update_config_dict(config, statement):
    """
    Use the given statement to apply changes to the given config dict
    """
    if not statement.is_assignment:
        return False
    strict = True  # (no option yet)
    try:
        dest, op = statement[:2]
        dest_dic = checked_reactor_dest(dest)
    except ValueError as e:
        if strict:
            raise
        print(e)
        return False
    else:
        rest = list(statement[2:])
        if not rest:
            raise ValueError('Incomplete assigment: %(statement)s'
                             % locals())
        key = dest_dic['key']
        subkey = dest_dic['subkey']
        key_factory = dest_dic['key_factory']
        tip_factory = dest_dic['tip_factory']
        tip_keys = tuple(dest_dic['tip_keys'])
        if rest[0].ttype != OP:
            valgrp = rest.pop(0)
            if rest and not rest[0].is_terminator:
                surplus = len(rest)
                raise ValueError('%(statement)r: nothing expected after '
                        '%(valgrp)r; %(surplus)d surplus token( group)s found!'
                        % locals())
            if subkey is None:
                if key_factory is dict:
                    raise ValueError('%(statement)r: missing subkey '
                            'for %(key)r; can\'t assign %(valgrp)r!'
                            % locals())
                set_value(config, key, valgrp, key_factory)
            else:
                if key_factory is None:
                    another_factory(dict, key)
                elif key_factory is not dict:
                    raise ValueError('%(statement)r: key_factory for (%(key)r,) '
                            'expected to be <dict>; found %(key_factory)r!'
                            % locals())
                subdict = config.setdefault(key, {})
                set_value(subdict, subkey, valgrp, tip_factory)
            return True

        if key_factory is dict:
            target_dict = config.setdefault(key, {})
        else:
            target_dict = config
            use_key = key
        pa1 = rest.pop(0)
        pa1tex = pa1.text
        if not pa1.opens_brace:
            raise ValueError('%(statement)r: opening brace expected; '
                             '%(pa1)r found!'
                             % locals())
        expected = pa1.expects_brace
        opens_dict = pa1tex == '{'
        if opens_dict:
            if subkey is not None:
                raise ValueError('%(statement)r: subkey %(subkey)r *and* '
                                 '%(pa1)r?!'
                                 % locals())
            elif tip_factory is None:
                another_factory(dict, *tip_keys)
                tip_factory = dict
            elif tip_factory is not dict:
                raise ValueError('%(statement)r: factory for %(tip_keys)s '
                        'expected to be <dict>; found %(tip_factory)r!'
                        % locals())
            more_names = 1
        else:
            # FINDOUT: do we need to distinguish lists and tuples?
            if pa1tex == '(':
                if tip_factory is None:
                    another_factory(tuple, *tip_keys)
                    tip_factory = tuple
                elif tip_factory is list:
                    pass
                elif tip_factory is not tuple:
                    raise ValueError('%(statement)r: factory for %(tip_keys)r '
                            'expected to be <tuple> (or list); '
                            'found %(tip_factory)r!'
                            % locals())
            else:
                assert pa1tex == '['
                if tip_factory is None:
                    another_factory(list, *tip_keys)
                    tip_factory = list
                elif tip_factory is tuple:
                    pass
                elif tip_factory is not list:
                    raise ValueError('%(statement)r: factory for %(tip_keys)r '
                            'expected to be <list> (or tuple); '
                            'found %(tip_factory)r!'
                            % locals())
            more_names = 0
        if more_names:
            name = None
            val = None
            for prev_tg, tg, next_tg in sequence_slide(rest):
                assert isinstance(tg, TokensGroup)
                ttype = tg.ttype
                text = tg.text
                if ttype == OP:
                    if text == ':':
                        if name is None:
                            raise ValueError('%(tg)r: expected to have a '
                                    'name here to assign to!'
                                    % locals())
                    elif tg.is_terminator:
                        assert next_tg is None
                        break
                    elif text == ',':
                        if name is None or val is None:
                            raise ValueError('%(tg)r: expected to have '
                                    'both a name and a value here; '
                                    'found %(name)r, %(val)r!'
                                    % locals())
                        name = None
                        val = None
                    elif text == expected:
                        if next_tg is None or next_tg.is_terminator:
                            pass
                        else:
                            raise ValueError('Unexpected %(next_tg)r '
                                    'following %(tg)r!'
                                    % locals())
                    else:
                        raise ValueError('Unexpected operator %(tg)r!'
                                % locals())
                elif name is None:
                    if ttype != NAME:
                        raise ValueError('%(tg)r: expected a name!'
                                % locals())
                    name = tg
                elif val is None:
                    val = tg
                    set_value(target_dict, name.text, val)
                else:
                    raise ValueError('%(tg)r: surplus value; '
                            'expected a comma or closing brace! '
                            '(name=%(name)r, val=%(val)r)'
                            % locals())
        else:
            factory = _factories.get(tip_keys + (None,))
            val = None
            vals = []
            for prev_tg, tg, next_tg in sequence_slide(rest):
                assert isinstance(tg, TokensGroup)
                ttype = tg.ttype
                text = tg.text
                if ttype == OP:
                    if tg.is_terminator:
                        assert next_tg is None
                        break
                    elif text in (',', expected):
                        if val is None:
                            raise ValueError('%(tg)r: expected to have '
                                    'a value here; '
                                    'found %(val)r!'
                                    % locals())
                        val = None
                    else:
                        raise ValueError('Unexpected operator %(tg)r!'
                                % locals())
                elif val is None:
                    val = tg
                    vals.append(resolve_value(val, factory))
                else:
                    raise ValueError('%(tg)r: surplus value; '
                            'expected a comma or closing brace! '
                            '(name=%(name)r, val=%(val)r)'
                            % locals())
            target_dict[use_key] = tip_factory(vals)


def amend_config_list(config, statement):
    """
    Use the given statement to append changes to the given config list
    """
    raise NotImplemented


if __name__ == '__main__':
    # if 0: (removed)
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: UTF-8 -*- äöü vim: ts=8 sts=4 sw=4 si et tw=79
"""\
tools.pdf.utils: Utility-Modul für den PDF-Export (mit PDFReactor)
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six.moves import map, range
from six.moves import StringIO

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

# Standard library:
from string import letters
from token import ENDMARKER, NAME, NEWLINE, OP, STRING, tok_name
from tokenize import (
    COMMENT,
    DEDENT,
    INDENT,
    NL,
    TokenError,
    generate_tokens,
    untokenize,
    )

# visaplan:
from visaplan.tools.minifuncs import makeBool
from visaplan.tools.minifuncs import translate_dummy as _
from visaplan.tools.sequences import sequence_slide

__all__ = ('ApiFilter',
           'acceptable_method_name',
           # ... verwendet: 'API_WHITELIST',
           )
# Logging / Debugging:
from visaplan.tools.debug import pp


def gen_restricted_lines(txt, func=None):
    """
    Generiere "Zeilen" aus dem übergebenen Text:

    >>> txt = '''# ein Kommentar
    ... strict on
    ... setAddLinks(False) # noch ein Kommentar
    ... with_images off
    ... '''

    >>> prefixed = make_prefixed('A')

    >>> list(gen_restricted_lines(txt, prefixed))
    ... # doctest: +NORMALIZE_WHITESPACE
    [<ControlStatement 'strict on '>,
     <ApiCall 'A.setAddLinks (False )'>,
     <ControlStatement 'with_images off '>]

    Einstmals hartcodierte Aufrufe:

    >>> txt2 = '''
    ... setAddLinks(False)
    ... # Enable bookmarks in the PDF document
    ... setAddBookmarks(True)
    ... setCleanupTool(CLEANUP_NONE)
    ... setEncoding('UTF-8')
    ... setJavaScriptMode(JAVASCRIPT_MODE_ENABLED_NO_LAYOUT)
    ... '''
    >>> list(gen_restricted_lines(txt2, prefixed))
    ... # doctest: +NORMALIZE_WHITESPACE
    [<ApiCall 'A.setAddLinks (False )'>,
     <ApiCall 'A.setAddBookmarks (True )'>,
     <ApiCall 'A.setCleanupTool (A.CLEANUP_NONE )'>,
     <ApiCall "A.setEncoding ('UTF-8')">,
     <ApiCall 'A.setJavaScriptMode (A.JAVASCRIPT_MODE_ENABLED_NO_LAYOUT )'>]

    Es können mehrere Anweisungen pro Zeile notiert werden,
    durch Semikolon getrennt:

    >>> txt3 = '; ;setA();setB()'
    >>> list(gen_restricted_lines(txt3, prefixed))
    [<ApiCall 'A.setA ()'>, <ApiCall 'A.setB ()'>]

    Nun zu den Neuerungen zur Unterstützung der neuen PDFreactor-API.
    Als Ersatz für die veralteten API-Aufrufe
    -- die (vorerst?) noch unterstützt und automatisch konvertiert werden --
    unterstützen wir nun direkte Zuweisungen an die config-Variable, die
    der .convert-Methode übergeben wird:

    >>> txt4 = '''
    ... config.outputFormat = {
    ...     type: OutputType.JPEG,
    ...     width: 640,
    ... }
    ... '''
    >>> list(gen_restricted_lines(txt4))      # doctest: +NORMALIZE_WHITESPACE
    [<Assignment 'config .outputFormat
                ={type :OutputType .JPEG ,width :640 ,}'>]
    """
    if not txt:
        return
    buf = []
    alltokens = generate_tokens(StringIO(txt or '').readline)
    for toktup in alltokens:
        ttype, token, startpos, endpos, line = toktup
        if ttype in (NEWLINE, ENDMARKER, DEDENT):
            if buf:
                yield restrictedStatement(buf, func)
                del buf[:]
        elif ttype == OP and token == ';':
            if buf:
                yield restrictedStatement(buf, func)
                del buf[:]
        elif ttype in (NL, COMMENT):
            pass
        elif not buf and ttype in (INDENT,
                ):
            pass
        else:
            buf.append(toktup)
    if buf:
        yield restrictedStatement(buf, func)

COMMAND_TYPES = list(range(100, 103))
CT_CONTROL, CT_ASSIGNMENT, CT_API = COMMAND_TYPES
"""\
toc (h2, h3, h4) afterbegin of body
appendix afterbegin of "#appendix" (
  images grouped force,
  media grouped auto,
  tables grouped auto,
  literature sorted,
  standards sorted
)"""

CONTROL_COMMANDS = ('strict',
                    'toc',
                    'appendix',
                    'with_images',
                    )
SYMBOLS_MAP = {}
_tmp = [('True',  'on', 'yes'),
        ('False', 'off', 'no'),
        ('None',  'null', 'nothing', 'nil'),
        ]
for _tup in _tmp:
    first = _tup[0]
    for name in _tup:
        SYMBOLS_MAP[name.lower()] = first

def make_prefixed(prefix):
    # copied from ./_parse.py 
    if not prefix:
        prefix = ''
    elif not prefix.endswith('.'):
        prefix += '.'
    def prefixed(token):
        if token[0] in '0123456789':
            int(token)
            return token
        ltoken = token.lower()
        if ltoken in SYMBOLS_MAP:
            return SYMBOLS_MAP[ltoken]
        else:
            return prefix + token
    return prefixed


class ApiError(ValueError):
    """
    Fehler bei Verarbeitung der PDFreactor-API-Aufrufe
    """
class BogusLine(ApiError):
    """
    Grundlegend fehlerhafte Zeile
    """
class DisallowedMethod(ApiError):
    """
    Versuch, eine unzulässige Methode aufzurufen
    """

def restrictedStatement(seq, transform_func):
    """
    Erzeuge aus der übergebenen (vom tokenize-Modul erzeugten) Sequenz einen
    API-Aufruf (Klasse ApiCall) oder eine Kontrollanweisung (ControlStatement).

    Einige Symbole werden normalisiert;
    so bestehen z. B. gewisse Freiheiten bei der Angabe logischer Werte:

    >>> func = make_prefixed('A')
    >>> repr(restrictedStatement([(1, 'strict'), (1, 'on')], func))
    "<ControlStatement 'strict on '>"
    >>> restrictedStatement([(1, 'strict'), (1, 'off')], func)
    <ControlStatement 'strict off '>

    Alle anderen Symbole werden, wie auch die Namen der API-Methoden,
    durch das Präfix <prefix> ergänzt:

    >>> restrictedStatement([(1, 'setDings'), (OP, '('),
    ...     (1, 'ATTRNAME'), (OP, ')')], func)
    <ApiCall 'A.setDings (A.ATTRNAME )'>

    (tokenize.untokenize fügt leider nach jedem Namen und jeder Zahl
    sicherheitshalber ein Leerzeichen ein)
    """
    res = []
    main = None
    ltype = None # line type, "Zeilen"typ
    for prevtup, currtup, nexttup in sequence_slide(seq):
        ttype, token = currtup[:2]
        WTF = tok_name[ttype]
        # pp(WTF, token)
        if main is None:
            if ttype == NAME:
                # set_trace()
                nexttype = (None
                            if nexttup is None
                            else nexttup[0])
                if nexttype is OP and nexttup[1] != '(':
                    ltype = CT_ASSIGNMENT
                elif token in CONTROL_COMMANDS:
                    ltype = CT_CONTROL
                elif nexttype == NAME:
                    # we might get an exception during untokenization :-(
                    # ... but at least it's an exception.
                    # We insist in the '.' OP after 'config'!
                    res.append(currtup)
                    raise BogusLine(untokenize(res))
                else:
                    ltype = CT_API
                main = token
        res.append(currtup)
    if ltype == CT_ASSIGNMENT:
        return Assignment(res, transform_func)
    elif ltype == CT_API:
        return ApiCall(res, transform_func)
    elif ltype == CT_CONTROL:
        return ControlStatement(res)
    elif not res:
        return None
    elif ltype is None:
        pp({'ltype:': ltype,
            'main': main,
            'res': res,
            })
        for toktup in res:
            token_info(*toktup)
        raise BogusLine(untokenize(res))
    else:
        raise ProgrammingError
    return ' '.join(res)

API_WHITELIST = frozenset(['enableDebugMode',
                           ])
def acceptable_method_name(name):
    """
    Gehört der übergebene Name zu einer akzeptablen PDFreactor-API-Funktion?
    """
    if name in API_WHITELIST:
        return True
    if not name[3:]:
        return False
    if (name.startswith('set')
        or name.startswith('add')
        ):
        return True
    return False

def token_info(ttype, token, start, end, logiline):
    """
    Für Entwicklung/Debugging
    """
    print('%-20s %r (%r)' % ('%2d (%s):' % (ttype, tok_name[ttype]),
                             token, logiline))

# ------------------------ [ Statements, generiert von ApiFilter ... [
class Statement(object):
    """
    Ein von der ApiFilter-Umgebung verwendbares Statement
    """

    def __init__(self, tokens, transform=None):
        tokheads = []
        name = None
        args = []
        has_errors = 0
        for toktup in tokens:
            ttype, token = toktup[:2]
            if name is None:
                if ttype == NAME:
                    name = token
                elif ttype in (INDENT,):
                    continue
                else:
                    if not has_errors:
                        pp(('Hu?', {
                            'tokens:': tokens,
                            'toktup:': toktup[:2],
                            }))
                    token_info(*toktup)
                    print('%s: Name erwartet' % (toktup,))
                    has_errors += 1
            elif ttype not in (OP,
                               ):
                args.append(token)
            # weitere Informationen (toktup[2:]) verwerfen:
            tokheads.append((ttype, token))
        self.tokens = tokheads
        self.name = name
        self.args = args
        if transform is not None:
            self.transformed = True
            tokens2 = []
            for toktup in tokens:
                ttype, token = toktup[:2]
                if ttype == NAME:
                    token = transform(token)
                tokens2.append((ttype, token))

            self.tokens_transformed = tokens2
        else:
            self.transformed = False

    def __str__(self):
        if self.transformed:
            return untokenize(self.tokens_transformed)
        else:
            return untokenize(self.tokens)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, str(self))

    def tell(self):
        pp({'name': self.name,
            'args': self.args,
            'tokens': self.tokens,
            'tokens_pretty': resolve_tokens(self.tokens),
            })

class ControlStatement(Statement):
    """
    Zur Steuerung der API-Aufrufe
    """

class ApiCall(Statement):
    """
    API-Aufruf; Methoden- und sonstige Namen werden mit einem Präfix versehen
    """

class Assignment(Statement):
    """
    Assignment of a value to a config key

    To support the new PDFreactor API, which uses a config dictionary instead
    of the plethora of setter methods, we support statements like

      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }

    which would be translated to something like

      dic = config.getdefault('outputFormat', {})
      dic.update({
          'type': PDFreactor.OutputType.JPEG,
          'width': 640,
          })

    """
# ------------------------ ] ... Statements, generiert von ApiFilter ]


# ------------------------------------------- [ Klasse ApiFilter ... [
class ApiFilter(object):
    """
    Erlaubt einen kontrollierten Zugriff auf eine API
    """

    def __init__(self,
                 config,
                 calls_before=None,
                 calls_default=None,
                 calls_after=None,
                 log=None,
                 msg=None,
                 check_methodname=None):
        """
        config -- the config dictionary to be given to the PDFreactor.convert
                  method; this is changed in-place.
        calls_before -- zu Beginn aufzurufende API-Aufrufe
                        (derzeit nicht verwendet)
        calls_default -- werden jeweils verwendet, sofern nicht dieselbe
                         Methode zuvor mit denselben *oder anderen* Argumenten
                         aufgerufen wurde
        calls_after -- abschließende API-Aufrufe; werden verwendet, sofern
                       nicht schon bei Abarbeitung der variablen Aufrufe
                       *identisch* erledigt (siehe __call__-Methode)
        symbols_map -- ein Dictionary, um z. B. Wahrheitswerte nach
                       Nicht-Python-Konventionen zu erkennen
        log -- eine Methode zum Loggen der Aufrufe
        msg -- der message-Adapter
        check_methodname -- eine Funktion, die API-Methodennamen daraufhin
                            prüft, ob ihre Verwendung erlaubt ist
        """
        self._config = config
        self.calls_before = calls_before
        self.calls_default = calls_default
        self.calls_after = calls_after
        self._log = log
        self._msg = msg
        self.check_methodname = check_methodname

    def log(self, *args, **kwargs):
        """
        Rufe die übergebene logger-Methode auf;
        ansonsten eine Notlösung
        """
        if self._log is not None:
            self._log(*args, **kwargs)
        else:
            print('#' * 79)
            print([(tup[0], repr(tup[1]))
                   for tup in kwargs.items()
                   ])
            print('#' * 79)
            print('(API)', ' '.join(map(repr, args)), \
                    ', '.join(['='.join((str(tup[0]), repr(tup[1]))
                                        )
                               for tup in kwargs.items()
                               ]))

    def msg(self, txt, type='info', mapping={}):
        """
        Rufe den Message-Adapter auf (wenn übergeben)
        """
        if self._msg is not None:
            self._msg(txt, type, mapping=mapping)

    def check_acceptable_name(self, name, **kwargs):
        """
        Der Name darf nur Buchstaben enthalten;
        außerdem sind einige Namen reserviert, um die korrekte Funktion
        sicherzustellen.

        """
        for ch in name:
            if not ch in letters:
                raise ValueError('Disallowed character %(ch)r'
                                 'in name %(name)r'
                                 % locals())
        if name in CONTROL_COMMANDS:
            raise ValueError('name %(name)r is reserved' % locals())
        if name in ('key', 'val', 'tup', 'prefix',
                    'statement', 'errors', 'found',
                    '_', 'translate', 'mogrify',
                    'text', 'errtext', 'ok', 'e',
                    'api_call',
                    # Zum Test dieser Funktionalität:
                    'verboten',
                    # Python-Standardfunktionen und -namen:
                    'str', 'isinstance',
                    'True', 'False', 'None', 'self',
                    ):
            raise ValueError('name %(name)r is reserved' % locals())

    def skip_this(self, stmt):
        """
        Gib True zurück, wenn sich Instanzen dieser ApiFilter-Klasse um das
        übergebene Statement prinzipiell nicht kümmern.
        """
        if isinstance(stmt, ControlStatement):
            return stmt.name != 'strict'
        elif isinstance(stmt, ApiCall):
            return stmt.name == 'setAddLinks'
        # was der Bauer nicht kennt ...:
        return True

    def _skip(self, stmt):
        if self.skip_this(stmt):
            self.log('%s: skipping %s' % (
                     self.__class__.__name__,
                     stmt))
            return True
        return False

    def _call_statement(self, stmt,
                        checkfunc=None):
        """
        Führe das übergebene Statement aus, sofern die bisherigen alle
        erfolgreich waren; andernfalls wird nur Versuch protokolliert, mit dem
        Vermerk 'Skip'.

        stmt -- ein Statement-Objekt
        checkfunc -- eine Funktion zur Prüfung, ob der übergebene Methodenname
                     zulässig ist. Wird nur beim Aufruf für die im Profil
                     konfigurierten, nicht für die hartcodierten Methoden
                     angegeben.

        Mit 'strict on' (Vorgabewert) ist der Versuch, eine unzulässige Methode
        aufzurufen, ein Fehler (und führt dazu, daß die weiteren nicht mehr
        ausgeführt werden).
        """
        if self._skip(stmt):
            return
        if isinstance(stmt, ApiCall):
            api_call = str(stmt)
            if not self._ctrl['ok']:
                errtext = _('Skipping %(api_call)r')
                self.log(errtext, locals())
                return
            if checkfunc is not None:
                if not checkfunc(stmt.name):
                    values = {'name': stmt.name,
                              }
                    self.msg('API method ${name} is disallowed', 'error', mapping=values)
                    self.log('API method %(name)r is disallowed', values)
                    if self._ctrl['strict']:
                        self._ctrl['ok'] = False
                        self.log('Remaining API calls will be skipped'
                                 '; PDF creation will be aborted')
                    return
            errtext = _('API call: %(api_call)s')
            self.log(errtext % locals())

            try:
                # erst das API-Objekt (den PDFreactor) lokal bekanntmachen:
                exec('%s=self._api_object' % self._api_name)
                # ... dann den Methodenaufruf:
                eval(api_call)
            except AttributeError as e:
                errtext = 'Unknown Attribute: %(e)r'
            except ValueError as e:
                errtext = 'Unknown Value: %(e)r'
            except NameError as e:
                errtext = 'Unknown Name: %(e)r'
            except Exception as e:
                e_name = e.__class__.__name__
                errtext = 'Unexpected Error (%(e_name)s): %(e)r'
            else:
                self.memorizeCall(stmt)
                return
            self.msg(_('API call ${api_call} failed'),
                     'error',
                     mapping=locals())
            self.log('E '+errtext, locals())
            self._ctrl['ok'] = False
            return
        elif isinstance(stmt, ControlStatement):
            if not self._ctrl['ok']:
                return
            # Bisher gibt es nur 'strict'.
            # Wenn es mehr werden, muß das evtl.
            # mal generischer gelöst werden:
            if stmt.name == 'strict':
                if stmt.args:
                    self._ctrl['strict'] = makeBool(stmt.args[0])
                else:
                    self._ctrl['strict'] = True
                self.log('i strict %s',
                         self._ctrl['strict'] and 'ON' or 'OFF')

    def memorizeCall(self, stmt):
        self._found.append((stmt.name, str(stmt)))

    def wasCalledBefore(self, stmt, exact):
        """
        Wurde das übergebene Statement zuvor schon aufgerufen?

        stmt -- das Statement
        exact -- Wahrheitswert
        """
        name = stmt.name
        filtered = [tup[1]
                    for tup in self._found
                    if tup[0] == name]
        if not exact:
            if filtered:
                self.log('Skipping %r: %r was called before',
                         str(stmt), name)
                return True
            else:
                return False
        else:
            s = str(stmt)
            if s in filtered:
                self.log('Skipping %r (was called before)',
                         s)
                return True
            elif filtered:
                self.log('Info: other calls of %r found',
                         name)
            return False

    def reset(self):
        self._found = []
        self._ctrl = {'strict': True,
                      'with_images': True,
                      'ok': True,
                      }

    def __call__(self, text,
                 **kwargs):
        """
        text -- zu interpretierender API-Text
        kwargs -- Es wird genau *ein* weiteres benanntes Argument erkannt;
                  dieses wird auch für das Präfix verwendet
        """
        if len(list(kwargs.keys())) != 1:
            raise ValueError('Expected exactly one named argument!'
                             ' (%s)' % kwargs)
        for tup in kwargs.items():
            key, val = tup
            self.check_acceptable_name(key)
            self._api_name = key
            self._api_object = val
            prefix = key + '.'
            break

        prefixed = make_prefixed(prefix)
        self.reset()

        checkfunc = self.check_methodname
        # text: Im Profil konfigurierte Aufrufe
        statements = list(gen_restricted_lines(text, prefixed))
        for statement in statements:
            self._call_statement(statement,
                                 checkfunc=checkfunc)
        # calls_default: Aufrufe von API-Methoden,
        # für die *keine* Aufrufe im Profil konfiguriert wurden
        for statement in gen_restricted_lines(self.calls_default, prefixed):
            if isinstance(statement, ControlStatement):
                # Kontrollanweisungen sind (vorerst?)
                # nur für konfigurierten Code gedacht:
                continue
            if not self.wasCalledBefore(statement, exact=False):
                self._call_statement(statement)
        # calls_after: Aufrufe von API-Methoden,
        #              die *exakt so* noch nicht erfolgt sind
        #              (aber nicht doppelt erfolgen sollen, wie z.B. das
        #              Hinzufügen von Skripten)
        for statement in gen_restricted_lines(self.calls_after, prefixed):
            if isinstance(statement, ControlStatement):
                # Kontrollanweisungen sind (vorerst?)
                # nur für konfigurierten Code gedacht:
                continue
            if not self.wasCalledBefore(statement, exact=True):
                self._call_statement(statement)

        return self._ctrl['ok']
# ------------------------------------------- ] ... Klasse ApiFilter ]


def fix(html, parser='lxml', from_encoding=None):
    """
    Gib eine geänderte und UTF-8-codierte Fassung des übergebenen HTML-Texts
    zurück.

    html -- der HTML-Text
    parser -- eine für BeautifulSoup 4 akzeptable Parser-Angabe
    from_encoding -- für den Fall, daß die automatische Erkennung versagt

    Literaturverweise werden geglättet und gesammelt:

    ><> uid = 'ccb9e2a26fc8ea71c567f0a39ddc4b85'
    ><> text = ('<html><body><p><a onclick="..." class="content-only" '
    ... 'href="./resolveuid/ccb9e2a26fc8ea71c567f0a39ddc4b85">[SteinD00]</a>')
    ><> fix(html)
    'DUMMY'
    """
    if from_encoding is not None:
        soup = BeautifulSoup(html, parser, from_encoding=from_encoding)
    else:
        soup = BeautifulSoup(html, parser)

    fix_literature(soup)
    return str(soup)


def fix_literature(soup):
    """
    Verarbeite Literaturlinks für die PDF-Ausgabe.

    soup -- das BeautifulSoup-Objekt, in-place geändert.
    """
    lit_list = []
    for a in soup.find_all('a', attrs={'class': 'content-only'}):
        text = a.get_text()
        if text.startswith('[') and text.endswith(']'):
            dic = {'text': text[1:-1],
                   }
            # TODO:
            # ... uid aus href extrahieren
            # ... onclick-Attribut entfernen
            # ... Änderungen (Element anhängen?), um Fußnoten zu erzeugen


"""<div class="attribute-text"><div class="clearfix layout-row columns-2"><div
class="layout-column first column-1"><p>Beim <strong>Abwasser</strong> handelt
es sich nach <a
onclick="window.open(this.href,'','resizable=yes,location=no,menubar=no,scrollbars=yes,status=no,toolbar=no,fullscreen=no,dependent=no,status,width=600,height=800');return
false;" href="./resolveuid/1749ea2e4c146dcfbf887e36dc14f357"
class="content-only BONK">[DINEN752g]</a> um ein Wasser, bestehend aus jeglicher Kombination von abgeleitetem Wasser aus Haushalten, Industrie- und Gewerbebetrieben, Oberflächenabfluss und unbeabsichtigtem Fremdwasserzufluss.</p></div> <div class="layout-column last column-2"><div style="text-align: left;">
 """

# ---------------------------------- [ Entwicklungsunterstützung ... [
def resolve_tokens(tokens):
    """
    Löse die numerischen Tokentypangaben auf.

    >>> tokens = [(1, 'toc'), (51, '('), (1, 'h1'), (51, ')')]
    >>> resolve_tokens(tokens)
    [('NAME', 'toc'), ('OP', '('), ('NAME', 'h1'), ('OP', ')')]
    """
    return [(tok_name[tup[0]], tup[1])
            for tup in tokens
            ]
# ---------------------------------- ] ... Entwicklungsunterstützung ]


if __name__ == '__main__':
  if 0:
      txt = '''
      strict on
      # don't forget that '.' OP:
      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      # Standard library:
      from pprint import pprint

      # Logging / Debugging:
      from pdb import set_trace
      print(list(gen_restricted_lines(txt)))


      bogustxt = '''
      strict on
      # don't forget that '.' OP:
      config outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      print(list(gen_restricted_lines(bogustxt)))

      txt = '''
      strict on
      config.outputFormat = {
          type: OutputType.JPEG,
          width: 640,
      }
      '''
      set_trace() # b 104
      for tokens in gen_restricted_lines(txt):
          pprint(tokens)
          # pprint(resolve_tokens(tokens))

  else:
    # Standard library:
    import doctest
    doctest.testmod()



"""
pdfreactor.plone.interfaces
"""

# Python compatibility:
from __future__ import absolute_import

from six import text_type as six_text_type

# Setup tools:
import pkg_resources

try:
    pkg_resources.get_distribution('pdfreactor.plone')
except pkg_resources.DistributionNotFound:
    HAVE_PDFREACTORPLONE = 0
else:
    HAVE_PDFREACTORPLONE = 1

# Zope:
from zope import schema
from zope.interface import Interface

# Plone:
from plone.supermodel import model

# Local imports:
from pdfreactor.parsecfg.defaults import default_config_text

if HAVE_PDFREACTORPLONE:
    from pdfreactor.plone.interfaces import \
        IGetPdfReactorConversionSettings as IGetSettings
else:
    # Zope:
    from Products.Five.browser import BrowserView
    class IGetSettings(BrowserView):
        """
        Get a config dict for PDFreactor.convert (e.a.)
        """


class IPdfReactorConversionSettings(model.Schema):
    """
    PDFreactor conversion configuration
    """
    config_text = schema.Text(
        title=u"PDFreactor default conversion settings",
        default=six_text_type(default_config_text),
        description=(u"Default configuration for PDFreactor exports on this "
        u"site; parsed and converted to a config dictionary as expected e.g. "
        u"by the PDFreactor.convert method."
        ))



# -*- coding: utf-8 -*-
# Python compatibility:
from __future__ import absolute_import

__author__ = """Tobias Herp <tobias.herp@visaplan.com>"""

# Zope:
from Missing import Value as MissingValue
from Products.CMFCore.utils import getToolByName
from zope.interface import implementer

# Plone:
from plone.app.upgrade.utils import loadMigrationProfile
from Products.CMFPlone.interfaces import INonInstallable

# visaplan:
from visaplan.plone.tools.setup import step

# Local imports:
from pdfreactor.parsecfg.defaults import default_config_text
from pdfreactor.parsecfg.plone.interfaces import IPdfReactorConversionSettings

# ------------------------------------------------------- [ data ... [
PROJECTNAME = 'pdfreactor.parsecfg'
PROFILE_ID = PROJECTNAME + ':default'
LOGGER_LABEL = PROJECTNAME + ': setuphandlers'
# ------------------------------------------------------- ] ... data ]

# Logging / Debugging:
import logging

logger = logging.getLogger(LOGGER_LABEL)


@implementer(INonInstallable)
class HiddenProfiles(object):

    def getNonInstallableProfiles(self):
        """Hide uninstall profile from site-creation and quickinstaller"""
        return [
            PROJECTNAME + ':uninstall',
        ]


def post_install(context):
    """Post install script"""
    logger.info('Installation complete')

    # Do something at the end of the installation of this package.


@step
def register_settings(context, logger):
    registry = getToolByName(context, 'portal_registry')

    registry.registerInterface(IPdfReactorConversionSettings)
    logger.info('Registered interface %r', IPdfReactorConversionSettings)
    proxy = registry.forInterface(IPdfReactorConversionSettings)
    key = 'config_text'
    early_return = 0
    late_return = 1
    if early_return:
        return
    val = getattr(proxy, key, None)
    if val not in (None, MissingValue):
        logger.info('Found %(key)s = %(val)r', locals())
    elif late_return:
        logger.warn('No value yet for key %(key)s!', locals())
    else:
        val = default_config_text.strip()
        setattr(proxy, key, val)
        logger.info('Set %(key)s to %(val)r', locals())



# Yes, we are aware of the "recommended approach for the highest level of
# compatibility"; see:
# https://packaging.python.org/en/latest/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
# ... but it didn't work for us, in our Zope / Plone environment;
# so for now, we stick with the solution which is as well used by Products.CMFPlone.
# See http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)
